"""

                       This is the __main__ module

This module calls the BioSANS function in BioSANS2020.RunBioSANS.

"""

# from subprocess import Popen
# from sys import executable
# Popen(str(executable)+" -m BioSANS2020.BioSANS")

from BioSANS2020.RunBioSANS import BioSANS
BioSANS()



"""

             This module is the convtopotosbml module

The sole purpose of this module is to faciliate conversion of topology
files into SBML files. Currently, this module still have a lot of
problems and the converted file may not run properly.

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from sympy import Symbol, Matrix
from libsbml import SBMLDocument, UNIT_KIND_ITEM, UNIT_KIND_SECOND, \
    UNIT_KIND_MOLE, UNIT_KIND_LITER, parseL3Formula, writeSBMLToFile

# from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.propagation.propensity import propensity_vec, \
    propensity_vec_molar
from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.propagation.recalculate_globals import get_globals


def topo_to_sbml(sp_comp, ks_vals, conc, r_react, r_prods, vbig, v_volms,
                 items=None, molar=False, rfile=None):
    """This function helps in the conversion of topology file to SBML.
    Consider a file containing the following reaction topology;

        #REACTIONS
        A <=> B, 0.5, 0.3
        3 B => 2 C, 0.04

        @CONCENTRATION
        A, 100
        B, 90
        C, 80

    Args:
        sp_comp : dictionary of components where the key is the
                  component and the value is a set of integers that tell
                  the rank or order of that reaction as it appears in
                  the file. For the above topology
                  sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}
        ks_vals : dictioary of parameters where the key is the line in
                  the file and the value is the list of parameter found
                  i.e. for the above topology;
                  ks_vals = {0: [0.5, 0.3], 1: [0.04]}
        conc : dictionary of initial concentration i.e.
               conc = {'A': 100.0, 'B': 90.0, 'C': 80}
        r_react : dictionary where the key is the line in the file and
                  the value is a dictionary of reactants where the key
                  is the component and the value is its stoichiometric
                  coefficient in that line i.e. for the above topology;
                  r_react = {0: {'A': 1}, 1: {'B': 3}}
        r_prods : dictionary where the key is the line in the file and
                  the value is a dictionary of products where the key
                  is the component and the value is its stoichiometric
                  coefficient in that line i.e. for the above topology;
                  r_prods = {0: {'B': 1}, 1: {'C': 2}}
        vbig : not used but pertains to stoichiometric matrix
        v_volms : volume of compartment
        items : 3 item list of [canvas, scroll_x, scroll_y]
        molar : True if conc. is in molar otherwise False
        rfile : path of the topology file + name of topology file
    """
    get_globals(rfile)
    cs_comp = {}
    for xvar in sp_comp:
        cs_comp[xvar] = Symbol(xvar, real=True, negative=False)

    kcs_vals = []
    for i, _ in enumerate(ks_vals):
        row = []
        if len(ks_vals[i]) == 1:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
        else:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
            key = 'kb' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
        kcs_vals.append(row)

    document = SBMLDocument(3, 1)
    model = document.createModel()
    model.setTimeUnits("second")

    comp = model.createCompartment()
    comp.setName('BasicCompartment')
    comp.setId('Basic')
    comp.setVolume(v_volms)
    comp.setUnits('litre')
    comp.setConstant(True)
    comp.setSpatialDimensions(3)

    substance = model.createUnitDefinition()
    substance.setId('substance')
    unit = substance.createUnit()
    unit.setKind(UNIT_KIND_ITEM)

    if not molar:
        fprop = Matrix(propensity_vec(kcs_vals, cs_comp, r_react, r_prods))
        units = []
        item_per_second = model.createUnitDefinition()
        item_per_second.setId('molecules_per_second')
        units.append(item_per_second.createUnit())
        units[-1].setKind(UNIT_KIND_ITEM)
        units[-1].setExponent(1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(item_per_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        per_second = model.createUnitDefinition()
        per_second.setId('per_second')
        units.append(per_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        per_item_second = model.createUnitDefinition()
        per_item_second.setId('per_molecules_second')
        units.append(per_item_second.createUnit())
        units[-1].setKind(UNIT_KIND_ITEM)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(per_item_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        molecules = model.createUnitDefinition()
        molecules.setId('molecules')
        units.append(molecules.createUnit())
        units[-1].setKind(UNIT_KIND_ITEM)
        units[-1].setExponent(1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        model.setSubstanceUnits("molecules")
    else:
        fprop = Matrix(
            propensity_vec_molar(
                kcs_vals,
                cs_comp,
                r_react,
                r_prods))
        units = []
        item_per_second = model.createUnitDefinition()
        item_per_second.setId('molar_per_second')
        units.append(item_per_second.createUnit())
        units[-1].setKind(UNIT_KIND_MOLE)
        units[-1].setExponent(1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(item_per_second.createUnit())
        units[-1].setKind(UNIT_KIND_LITER)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(item_per_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        per_second = model.createUnitDefinition()
        per_second.setId('per_second')
        units.append(per_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        per_item_second = model.createUnitDefinition()
        per_item_second.setId('per_molar_second')
        units.append(per_item_second.createUnit())
        units[-1].setKind(UNIT_KIND_LITER)
        units[-1].setExponent(1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(per_item_second.createUnit())
        units[-1].setKind(UNIT_KIND_MOLE)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(per_item_second.createUnit())
        units[-1].setKind(UNIT_KIND_SECOND)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)

        molarity = model.createUnitDefinition()
        molarity.setId('moles_per_liter')
        units.append(molarity.createUnit())
        units[-1].setKind(UNIT_KIND_MOLE)
        units[-1].setExponent(1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        units.append(molarity.createUnit())
        units[-1].setKind(UNIT_KIND_LITER)
        units[-1].setExponent(-1)
        units[-1].setScale(0)
        units[-1].setMultiplier(1)
        model.setSubstanceUnits("moles_per_liter")

    spcs = []
    for sp_i in sp_comp:
        spcs.append(model.createSpecies())
        spcs[-1].setName(sp_i)
        spcs[-1].setId(sp_i)
        spcs[-1].setCompartment('Basic')
        spcs[-1].setInitialAmount(float(conc[sp_i]))
        # if molar == False:
        if not molar:
            spcs[-1].setSubstanceUnits('molecules')
        # elif molar == True:
        elif molar:
            spcs[-1].setSubstanceUnits('moles_per_liter')
        else:
            spcs[-1].setHasOnlySubstanceUnits(True)

    rcks_var = []
    for i, _ in enumerate(kcs_vals):
        if len(kcs_vals[i]) == 1:
            rcks_var.append(model.createParameter())
            rcks_var[-1].setId(str(kcs_vals[i][0]))
            rcks_var[-1].setConstant(True)
            rcks_var[-1].setValue(ks_vals[i][0])
            dvar = sum([r_react[i][x] for x in r_react[i]])
            # if molar == False:
            if not molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molecules_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molecules_second')
            # elif molar == True:
            elif molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molar_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molar_second')
            else:
                pass
            rcks_var[-1].setConstant(True)
        else:
            rcks_var.append(model.createParameter())
            rcks_var[-1].setId(str(kcs_vals[i][0]))
            rcks_var[-1].setConstant(True)
            rcks_var[-1].setValue(ks_vals[i][0])
            dvar = sum([r_react[i][x] for x in r_react[i]])
            # if molar == False:
            if not molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molecules_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molecules_second')
            # elif molar == True:
            elif molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molar_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molar_second')
            else:
                pass
            rcks_var[-1].setConstant(True)

            rcks_var.append(model.createParameter())
            rcks_var[-1].setId(str(kcs_vals[i][1]))
            rcks_var[-1].setConstant(True)
            rcks_var[-1].setValue(ks_vals[i][1])
            dvar = sum([r_react[i][x] for x in r_react[i]])
            # if molar == False:
            if not molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molecules_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molecules_second')
            # elif molar == True:
            elif molar:
                if dvar == 0:
                    rcks_var[-1].setUnits('molar_per_second')
                elif dvar == 1:
                    rcks_var[-1].setUnits('per_second')
                elif dvar == 2:
                    rcks_var[-1].setUnits('per_molar_second')
            else:
                pass
            rcks_var[-1].setConstant(True)

    rxns = []
    for rvar in r_react:
        rxns.append(model.createReaction())
        rxns[-1].setId("Rf" + str(int(rvar) + 1))
        rxns[-1].setReversible(False)
        rbig_var = []
        for avar in r_react[rvar]:
            if int(r_react[rvar][avar]) > 0 and avar[0] != "-":
                rbig_var.append(rxns[-1].createReactant())
                rbig_var[-1].setSpecies(avar)
                rbig_var[-1].setStoichiometry(r_react[rvar][avar])

        pbig_var = []
        for bvar in r_prods[rvar]:
            if int(r_prods[rvar][bvar]) > 0 and bvar[0] != "-":
                pbig_var.append(rxns[-1].createProduct())
                pbig_var[-1].setSpecies(bvar)
                pbig_var[-1].setStoichiometry(r_prods[rvar][bvar])
        if len(ks_vals[rvar]) == 2:
            rxns.append(model.createReaction())
            rxns[-1].setId("Rb" + str(int(rvar) + 1))
            rxns[-1].setReversible(False)
            rbig_var = []
            for avar in r_react[rvar]:
                if int(r_react[rvar][avar]) > 0 and avar[0] != "-":
                    rbig_var.append(rxns[-1].createProduct())
                    rbig_var[-1].setSpecies(avar)
                    rbig_var[-1].setStoichiometry(r_react[rvar][avar])

            pbig_var = []
            for bvar in r_prods[rvar]:
                if int(r_prods[rvar][bvar]) > 0 and bvar[0] != "-":
                    pbig_var.append(rxns[-1].createReactant())
                    pbig_var[-1].setSpecies(bvar)
                    pbig_var[-1].setStoichiometry(r_prods[rvar][bvar])

    kinetic_law = []
    for rvar, _ in enumerate(rxns):
        kinetic_law.append(rxns[rvar].createKineticLaw())
        kinetic_law[-1].setMath(parseL3Formula(str(fprop[rvar]
                                                   ).replace("**", "^")))

    document.setModel(model)
    writeSBMLToFile(document, globals2.TO_CONVERT + ".xml")
    print(vbig, items)
    return [0, 0]



"""

            This module is the law_of_localization module

The sole purpose of this module is to implement the law of localization


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.propagation.propensity import propensity_vec, \
    propensity_vec_molar
from sympy import Matrix, Symbol, simplify, NonSquareMatrixError


def subs2(zvar, cval):
    """This function substitute actual values to sympy symbols.

    Args:
        zvar (sympy.core): sympy expression
        cval (disct): dictionary of Symbols : value

    Returns:
        (sympy.core): substituted expression
    """
    for xvar in cval:
        zvar = zvar.subs(xvar, cval[xvar])
    return zvar


def prepare_ffrrint(items):
    """This function instantiate how the output will be printed by crea-
    thing the function ffprint

    Args:
        text (Text): text area where the outputs are written
        items (tuple): 3 item list of (canvas, scroll_x, scroll_y)

    Returns:
        function: either print in console or in text area
    """
    if items:
        text = prepare_scroll_text(items)

        def ffprint(xvar):
            return text.insert('insert', " ".join([str(y) for y in xvar]))
    else:

        def ffprint(xvar):
            return print(" ".join([str(y) for y in xvar]), end="")
    return ffprint


def prepare_dict_list1(sp_comp, conc, ks_dict):
    """This function prepares cs_var, cso_var, time

    Args:
        sp_comp (dict): dictionary of components position in reaction
        conc (dict) : dictionary of initail concentration

    Returns:
        (dict/list): cs_var, cso_var, equivals, equi_ks
    """
    cs_var = {}
    cso_var = {}
    equivals = []
    equi_co = []
    equi_ks = []

    for xvar in sp_comp:
        cs_var[xvar] = Symbol(xvar, real=True, negative=False)
        cso_var[xvar] = Symbol(xvar + 'o', real=True, negative=False) * \
            (0 if conc[xvar] == 0 else 1)
        equivals.append((cso_var[xvar], conc[xvar]))
        equi_co.append((cso_var[xvar], conc[xvar]))

    kcs = []
    for i, _ in enumerate(ks_dict):
        row = []
        if len(ks_dict[i]) == 1:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[0], ks_dict[i][0]))
            equi_ks.append((row[0], ks_dict[i][0]))
        else:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[0], ks_dict[i][0]))
            equi_ks.append((row[0], ks_dict[i][0]))
            key = 'kb' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[1], ks_dict[i][1]))
            equi_ks.append((row[1], ks_dict[i][1]))
        kcs.append(row)
    return cs_var, equivals, equi_co, equi_ks, kcs


def prepare_list2(stch_var, slabels):
    """This function prepares two list from stoichiometric matrix

    Args:
        stch_var (np.ndarray): toichiometric matrix

    Returns:
        list: [non zero row values of  stch_var,
            nonzero rows index of stch_var]
    """
    ss_var = []
    nz_var = []
    for row in range(stch_var.shape[0]):
        if sum(abs(stch_var[row, :])) != 0 and slabels[row][0] != "-":
            ss_var.append(list(stch_var[row, :]))
            nz_var.append(row)
    ss_var = Matrix(ss_var)
    return ss_var, nz_var


def law_loc_symbolic(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                     items=None, molar=False, mode=None, numer=False):
    """[summary]

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (np.ndarray): stoichiometric matrix
        items (tuple): 3 item list of (canvas, scroll_x, scroll_y)
        molar (bool, optional): [description]. Defaults to False.
        mode (str, optional): either "Numeric", "fofks", "fofCo".
        numer (bool, optional): If True, numeric substitution will be
            done. Defaults to False.

    Returns:
        list: [0, 0] - useless return
    """

    ffprint = prepare_ffrrint(items)
    cs_var, equivals, equi_co, equi_ks, kcs \
        = prepare_dict_list1(sp_comp, conc, ks_dict)

    if not molar:
        prop_flux = Matrix(propensity_vec(kcs, cs_var, r_dict, p_dict))
    else:
        prop_flux = Matrix(propensity_vec_molar(kcs, cs_var, r_dict, p_dict))

    if mode == "Numeric":
        prop_flux = prop_flux.subs(equivals)
    elif mode == "fofks":
        prop_flux = prop_flux.subs(equi_co)
    elif mode == "fofCo":
        prop_flux = prop_flux.subs(equi_ks)

    stch_var = Matrix(stch_var)
    # cs_var might have change after call
    for xvar in sp_comp:
        cs_var[xvar] = Symbol(xvar, real=True, negative=False)
    slabels = list(cs_var.keys())

    ss_var, nz_var = prepare_list2(stch_var, slabels)

    ccs = [cs_var[xvar] for xvar in cs_var]
    js_name = [ccs[xvar] for xvar in nz_var]
    a_jac = prop_flux.jacobian(js_name)
    ker_v = ss_var.nullspace()

    as_jac = a_jac.col_insert(a_jac.shape[0], -ker_v[0])
    for ih_index in range(1, len(ker_v)):
        as_jac = as_jac.col_insert(as_jac.shape[0], -ker_v[ih_index])

    knames = []
    for i, _ in enumerate(kcs):
        if len(kcs[i]) == 1:
            knames.append(kcs[i][0])
        else:
            knames.append(kcs[i][0])
            knames.append(kcs[i][1])

    ffprint([
        "For inhibition or downregulation : \n\nWhen k decrease, the \
            sign of the sensitivity tells what happen to species activity\n"
    ])
    ffprint([
        "When k increase, the reverse of the sign of the sensitivity \
            tells what happen to species activity\n\n"
    ])

    if numer:
        as_jac = as_jac.subs(equivals)

    try:
        stch_var = simplify(as_jac.inv()).T
        if numer:
            stch_var = stch_var.subs(equivals)
        for ih_index in range(stch_var.shape[0]):
            for ij_index in range(stch_var.shape[1]):
                try:
                    ffprint([knames[ih_index], "\t", js_name[ij_index],
                             "\t", stch_var[ih_index, ij_index], "\n"])
                except:
                    pass
            ffprint(["\n"])
    except NonSquareMatrixError:
        ffprint(["Non square matrix error"])
    except:
        pass

    return [0, 0]



"""

                 This module is the runge_kutta4 module

This module serves in the integration of ODE trajectory using RK4 and
tau-adaptive RK4-algorithm.

The list of functions in this module are

1. rk4_model
2. runge_kutta_forth
3. rungek4_int
rkck
rkqs
rungek4a_int


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.propagation.propensity import propensity_vec, \
    propensity_vec_molar
from BioSANS2020.propagation.recalculate_globals import get_globals, \
    apply_rules
from BioSANS2020.myglobal import mglobals as globals2


def rk4_model(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
              tvar, molar=False):
    """This function returns  the evaluated  value of the  derivative of
    each component with respwct to time at a particular instant based on
    the state of the system at that instant.

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ksn_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stoch_var (numpy.ndarray): stoichiometric matrix. For example

            stoch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        tvar (list): time stamp of trajectories i.e. [0, 0.1, 0.2, ...]
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.

    Returns:
        np.ndarray: value of dx/dt
    """
    if not molar:
        prop_flux = propensity_vec(ks_dict, conc, r_dict, p_dict)
    else:
        prop_flux = propensity_vec_molar(ks_dict, conc, r_dict, p_dict)
    dxdt = np.matmul(stch_var, prop_flux).reshape(len(sp_comp))
    if globals2.CON_BOUNDARY:
        spc = list(sp_comp.keys())  # [spi for spi in sp_comp]
        for xvar in globals2.CON_BOUNDARY:
            ind = spc.index(xvar)
            dxdt[ind] = 0*tvar[0]  # useless multiplication
    return dxdt


def runge_kutta_forth(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                      tvar, delt, n_sp, molar=False):
    """this function prepare some stuffs for the integration

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc ([type]): [description]
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }

        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        v_stoich (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        tvar (list): time stamp of trajectories i.e. [0, 0.1, 0.2, ...]
        delt (float): step size
        n_sp (dict): dictionary of keywords
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.

    Returns:
        list: [updated concn, updated time]
    """

    yconc = {xvar: conc[xvar] for xvar in n_sp}

    da_dt = rk4_model(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                      tvar, molar)
    new_a1 = da_dt * delt
    ind = 0
    for spi in sp_comp:
        yconc[spi] = conc[spi] + 0.5 * new_a1[ind]
        ind = ind + 1

    da_dt = rk4_model(sp_comp, ks_dict, yconc, r_dict, p_dict, stch_var,
                      tvar + 0.5 * delt, molar)
    new_a2 = da_dt * delt
    ind = 0
    for spi in sp_comp:
        yconc[spi] = conc[spi] + 0.5 * new_a2[ind]
        ind = ind + 1

    da_dt = rk4_model(sp_comp, ks_dict, yconc, r_dict, p_dict, stch_var,
                      tvar + 0.5 * delt, molar)
    new_a3 = da_dt * delt
    ind = 0
    for spi in sp_comp:
        yconc[spi] = conc[spi] + 0.5 * new_a3[ind]
        ind = ind + 1

    da_dt = rk4_model(sp_comp, ks_dict, yconc, r_dict, p_dict, stch_var,
                      tvar + delt, molar)
    new_a4 = da_dt * delt

    new = (new_a1 + 2.0 * new_a2 + 2.0 * new_a3 + new_a4) / 6.0

    new_a = yconc
    ind = 0
    for spi in sp_comp:
        new_a[spi] = conc[spi] + new[ind]
        ind = ind + 1

    return [new_a, tvar + delt]


def rungek4_int(conc, time, sp_comp, ks_dict, r_dict, p_dict, stch_var,
                molar=False, delx=1, rfile=""):
    """Peforms the RK4 integration

    Args:
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        tvar (list): time stamp of trajectories i.e. [0, 0.1, 0.2, ...]
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.
        delx (float, optional): stepsize modifier. Defaults to 1.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        list: [time stamp, trajectory]
    """
    get_globals(rfile)
    tend = time[-1]
    div = max(1, int(1 / delx))
    delt = (time[-1] - time[-2]) / div
    yconc = {xvar: conc[xvar] for xvar in conc}
    slabels = list(sp_comp.keys())  # [a for a in sp_comp]
    apply_rules(conc, yconc, [0], [conc[a] for a in sp_comp], slabels)
    zvar = [conc[a] for a in sp_comp]
    zlist = [zvar]
    tvar = 0.0
    t_c = [tvar]
    n_sp = [xvar for xvar in conc if xvar not in sp_comp]
    d_v = 0
    tc2 = [tvar]
    z_2 = [zvar]

    while abs(tvar - tend) > 1.0e-10:
        conc, tvar = runge_kutta_forth(
            sp_comp, ks_dict, conc, r_dict, p_dict, stch_var, tvar,
            delt, n_sp, molar)
        # conc, yerr = rkck(delt,sp_comp,ks_dict,conc,r_dict,p_dict,
        # stch_var,tvar,n_sp,molar) #higher order runge-kutta
        # tvar = tvar + delt     # higher order
        # runge-kutta
        apply_rules(conc, yconc, tc2, z_2, slabels)
        z_2.append([conc[a] for a in sp_comp])
        tc2.append(tvar)
        if d_v == div - 1:
            zlist.append([conc[a] for a in sp_comp])
            t_c.append(tvar)
            d_v = 0
        else:
            d_v = d_v + 1
    zlist = np.array(zlist)
    return [t_c, zlist]


def rkck(hvar, sp_comp, ks_dict, conc, r_dict, p_dict,
         stch_var, tvar, n_sp, molar):
    """This function is a helper function for tau-adaptive RK4"""

    a_2 = 0.2
    a_3 = 0.3
    a_4 = 0.6
    a_5 = 1.0
    a_6 = 0.875

    b_21 = 0.2
    b_31 = 3.0 / 40.0
    b_32 = 9.0 / 40.0

    b_41 = 0.3
    b_42 = -0.9
    b_43 = 1.2

    b_51 = -11.9 / 54.0
    b_52 = 2.5
    b_53 = -70.0 / 27.0
    b_54 = 35.0 / 27.0

    b_61 = 1631.0 / 55296.0
    b_62 = 175.0 / 512.0
    b_63 = 575.0 / 13824.0
    b_64 = 44275.0 / 110592.0
    b_65 = 253.0 / 4096.0

    c_1 = 37.0 / 378.0
    c_3 = 250.0 / 621.0
    c_4 = 125.0 / 594.0
    c_6 = 512.0 / 1771.0

    dc_1 = c_1 - 2825.0 / 27648.0
    dc_3 = c_3 - 18575.0 / 48384.0
    dc_4 = c_4 - 13525.0 / 55296.0
    dc_5 = -277.0 / 14336.0
    dc_6 = c_6 - 0.25

    dydx = rk4_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                     stch_var, tvar, molar)

    ytemp = {xvar: conc[xvar] for xvar in n_sp}
    ind = 0
    for spi in sp_comp:
        ytemp[spi] = conc[spi] + b_21 * hvar * dydx[ind]
        ind = ind + 1
    ak2 = rk4_model(sp_comp, ks_dict, ytemp, r_dict, p_dict,
                    stch_var, tvar + a_2 * hvar, molar)

    ind = 0
    for spi in sp_comp:
        ytemp[spi] = conc[spi] + hvar * (b_31 * dydx[ind] + b_32 * ak2[ind])
        ind = ind + 1
    ak3 = rk4_model(sp_comp, ks_dict, ytemp, r_dict, p_dict,
                    stch_var, tvar + a_3 * hvar, molar)

    ind = 0
    for spi in sp_comp:
        ytemp[spi] = conc[spi] + hvar * \
            (b_41 * dydx[ind] + b_42 * ak2[ind] + b_43 * ak3[ind])
        ind = ind + 1
    ak4 = rk4_model(sp_comp, ks_dict, ytemp, r_dict, p_dict,
                    stch_var, tvar + a_4 * hvar, molar)

    ind = 0
    for spi in sp_comp:
        ytemp[spi] = conc[spi] + hvar * \
            (b_51 * dydx[ind] + b_52 * ak2[ind] +
             b_53 * ak3[ind] + b_54 * ak4[ind])
        ind = ind + 1
    ak5 = rk4_model(sp_comp, ks_dict, ytemp, r_dict, p_dict,
                    stch_var, tvar + a_5 * hvar, molar)

    ind = 0
    for spi in sp_comp:
        ytemp[spi] = conc[spi] + hvar * (b_61 * dydx[ind] + b_62
                                         * ak2[ind] + b_63 * ak3[ind]
                                         + b_64 * ak4[ind] + b_65 * ak5[ind])
        ind = ind + 1
    ak6 = rk4_model(sp_comp, ks_dict, ytemp, r_dict, p_dict,
                    stch_var, tvar + a_6 * hvar, molar)

    yout = ytemp
    yerr = []
    ind = 0
    for spi in sp_comp:
        yout[spi] = conc[spi] + hvar * \
            (c_1 * dydx[ind] + c_3 * ak3[ind] +
             c_4 * ak4[ind] + c_6 * ak6[ind])
        yerr.append(hvar * (dc_1 * dydx[ind] + dc_3 * ak3[ind] + dc_4 *
                            ak4[ind] + dc_5 * ak5[ind] + dc_6 * ak6[ind]))
        ind = ind + 1

    return [yout, yerr]


def rkqs(htry, eps, yscal, sp_comp, ks_dict, conc, r_dict, p_dict,
         stch_var, tvar, n_sp, molar):
    """This function is a helper function for tau-adaptive RK4"""
    safety = 0.9
    pgrow = -0.2
    pshrnk = -0.25
    errcon = 1.89e-4
    errmax = 1000
    hvar = htry
    while errmax > 1:
        ytemp, yerr = rkck(hvar, sp_comp, ks_dict, conc, r_dict, p_dict,
                           stch_var, tvar, n_sp, molar)
        errmax = max(0, np.max(np.array(yerr) / np.array(yscal))) / eps
        if errmax > 1:
            hvar = safety * hvar * (errmax**pshrnk)
            if hvar < 0.1 * hvar:
                hvar = 0.1 * hvar
        else:
            if errmax > errcon:
                hnext = safety * hvar * (errmax**pgrow)
            else:
                hnext = 5.0 * hvar
            hdid = hvar
            conc = ytemp

    return [conc, hdid, hnext, yerr]


def rungek4a_int(tvar, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                 yscal=10, molar=False, implicit=False, rfile=""):
    """[summary]

    Args:
        tvar (list): time stamp of trajectories i.e. [0, 0.1, 0.2, ...]
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        yscal (int, optional): [description]. Defaults to 10.
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        list: [time stamp, trajectory]
    """
    get_globals(rfile)
    tnew = []
    delt = tvar[-1] - tvar[-2]
    eps = 1.0e-8
    n_sp = [xvar for xvar in conc if xvar not in sp_comp]
    yconc = {xvar: conc[xvar] for xvar in sp_comp}
    slabels = list(sp_comp.keys())  # [a for a in sp_comp]
    apply_rules(conc, yconc, [0], [conc[a] for a in sp_comp], slabels)
    yvar = [conc[a] for a in sp_comp]
    s_list = [yvar]
    if not implicit:
        tnow = tvar[0]
        tnew.append(tnow)
        tindex = 1

        while abs((tnew[-1] - tvar[-1]) / tvar[-1]) > 1.0e-5:

            y_old = conc
            conc, dtime, delt, _ = rkqs(
                delt, eps, yscal, sp_comp, ks_dict, conc, r_dict,
                p_dict, stch_var, tnow, n_sp, molar)
            tnow = tnow + dtime
            # apply_rules(conc, yconc)
            apply_rules(conc, yconc, tnew, s_list, slabels)
            if tnow > tvar[tindex]:
                tnow = tnow - dtime
                dtime = tvar[tindex] - tnow
                conc, tnow = runge_kutta_forth(
                    sp_comp, ks_dict, y_old, r_dict, p_dict, stch_var,
                    tnow, dtime, n_sp, molar)
                delt = 5 * dtime
                # apply_rules(conc, yconc)
                apply_rules(conc, yconc, tnew, s_list, slabels)
                tindex = tindex + 1
            s_list.append([conc[a] for a in sp_comp])
            tnew.append(tnow)
    else:
        tnow = tvar[0]
        tnew.append(tnow)
        tindex = 1

        while abs((tnew[-1] - tvar[-1]) / tvar[-1]) > 1.0e-5:

            y_old = conc
            conc, dtime, delt, _ = rkqs(
                delt, eps, yscal, sp_comp, ks_dict, conc, r_dict, p_dict,
                stch_var, tnow, n_sp, molar)
            tnow = tnow + dtime
            # apply_rules(conc, yconc)
            apply_rules(conc, yconc, tnew, s_list, slabels)
            if tnow > tvar[tindex]:
                tnow = tnow - dtime
                dtime = tvar[tindex] - tnow
                conc, tnow = runge_kutta_forth(
                    sp_comp, ks_dict, y_old, r_dict, p_dict, stch_var,
                    tnow, dtime, n_sp, molar)
                delt = 5 * dtime
                # apply_rules(conc, yconc)
                apply_rules(conc, yconc, tnew, s_list, slabels)
                s_list.append([conc[a] for a in sp_comp])
                tnew.append(tnow)
                tindex = tindex + 1
    return [tnew, np.array(s_list)]



"""

                  This module is the tau_leaping2 module

The purpose of this module is to propagate stochastic trajectories using
the tau-leaping algorithm.

The functions in this module are the following;

1. tau_leaping2
2. step_3to5
3. ssa_support


"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.propagation.propensity import propensity_vec
from BioSANS2020.propagation.recalculate_globals \
    import get_globals, apply_rules, reserve_events_words


def tau_leaping2(tvar, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                 rand_seed, del_coef=1, implicit=False, rfile=""):
    """This functions performs the tau-leaping integration

    Args:
        tvar ([type]): [description]
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
         p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        rand_seed ([type]): [description]
        rand_seed (float): random  seed value picked  at random for each
            trajectory. They have been sampled from the calling program.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple : (time, trajectories)
    """
    get_globals(rfile)
    tmax = tvar[-1]
    np.random.seed(int(rand_seed * 100))
    tnew = []
    stch_var2 = stch_var**2
    stch_var = stch_var.T
    dto = tvar[-1] - tvar[-2]
    tchlen = len(globals2.TCHECK)

    all_sp = list(sp_comp.keys())  # [z for z in sp_comp]
    spc = [z for z in sp_comp if z not in reserve_events_words]
    spc2 = [z for z in sp_comp if z in reserve_events_words]
    concz = {xvar: conc[xvar] for xvar in conc}
    yconc = {xvar: conc[xvar] for xvar in conc}
    apply_rules(concz, yconc)
    update_sp = [all_sp.index(z) for z in spc]

    vcri = np.where(stch_var < 0)
    vncr = np.array(list(range(len(stch_var[0]))))
    vncr = vncr[~np.isin(vncr, vcri[1])]

    zconc = [[concz[z] for z in spc]]
    t_c = 0
    tnew.append(t_c)

    g_i = []
    for i, _ in enumerate(spc):
        keys = sp_comp[spc[i]]
        maxrlen = 0
        rlen = 0
        for key in keys:
            if spc[i] in r_dict[key]:
                maxrlen = max(maxrlen, len(r_dict[key]))
            elif spc[i] in p_dict[key]:
                maxrlen = max(maxrlen, len(p_dict[key]))
        if maxrlen == 1:
            for key in keys:
                if spc[i] in r_dict[key]:
                    rlen = max(rlen, r_dict[key][spc[i]])
                elif spc[i] in p_dict[key]:
                    rlen = max(rlen, p_dict[key][spc[i]])
            if rlen == 1:
                g_i.append(lambda xvar: 1)
            else:
                g_i.append(lambda xvar: 2)
        elif maxrlen == 2:
            try:
                g_i.append(lambda xvar: 2 + 1 / (xvar - 1))
            except:
                g_i.append(lambda xvar: 2)
        else:
            g_i.append(lambda xvar: 1)

    if not implicit:
        while t_c < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            if len(vcri[1]): # > 0:
                lcri = set()
                lncr = set()
                for xvar in range(len(vncr)):
                    lncr.add(xvar)

                for xvar in range(len(vcri[0])):
                    i, j = [vcri[0][xvar], vcri[1][xvar]]
                    if abs(zconc[-1][i] / stch_var[i, j]
                           ) < 10 and prop_flux[j] > 0:
                        lcri.add(j)
                    else:
                        lncr.add(j)
            else:
                lcri = set()

            if len(lcri) == len(stch_var[0]):
                alp = np.sum(prop_flux)
                d_t = (1 / alp) * (np.log(1 / np.random.uniform()))
            elif not lcri:  # len(lcri) == 0:
                alp = np.sum(prop_flux)
                uuj = np.matmul(stch_var, prop_flux)
                sig = np.matmul(stch_var2, prop_flux)
                exigi = np.array([zconc[-1][j] * (1 / g_i[j](zconc[-1][j]))
                                  for j in range(len(spc))]) * 0.03 * del_coef
                exigi1 = np.maximum(exigi, np.full(len(spc), 1))
                d_t = min(np.min(exigi1 / np.abs(uuj)),
                          np.min(exigi1 * exigi1 / np.abs(sig)))
            else:
                lcri = np.array(list(lcri))
                lncr = np.array(list(lncr))
                alpc = np.sum(prop_flux[lcri])
                dtc = (1 / alpc) * (np.log(1 / np.random.uniform()))

                alp = np.sum(prop_flux[lncr])
                uuj = np.matmul(stch_var[:, lncr], prop_flux[lncr])
                sig = np.matmul(stch_var2[:, lncr], prop_flux[lncr])
                exigi = np.array([zconc[-1][j] * (1 / g_i[j](zconc[-1][j]))
                                  for j in range(len(spc))]) * 0.03 * del_coef
                exigi1 = np.maximum(exigi, np.full(len(spc), 1))
                d_t = min(np.min(exigi1 / np.abs(uuj)),
                          np.min(exigi1 * exigi1 / np.abs(sig)), dtc)

            kvar = np.random.poisson(prop_flux * d_t)
            all_pos = True
            c_c = {}
            b_b = np.sum(kvar * stch_var[:, update_sp], 0)
            for xvar, _ in enumerate(spc):
                holder = zconc[-1][xvar] + b_b[xvar]
                if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                    c_c[spc[xvar]] = holder
                else:
                    all_pos = False
                    break
            if all_pos:
                for xvar, _ in enumerate(spc):
                    concz[spc[xvar]] = c_c[spc[xvar]]
                for xvar, _ in enumerate(spc2):
                    concz[spc2[xvar]] = concz[spc2[xvar]] + d_t
                apply_rules(concz, yconc)
                zconc.append([concz[xvar] for xvar in spc])
                t_c = t_c + d_t
                tnew.append(t_c)
    else:
        tindex = 1
        index = 0
        cvar = [concz[z] for z in spc]
        while tvar[tindex] < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            # step 1
            if len(vcri[1]):  # > 0:
                lcri = set()
                lncr = set()
                for xvar in range(len(vncr)):
                    lncr.add(xvar)

                for xvar in range(len(vcri[0])):
                    i, j = [vcri[0][xvar], vcri[1][xvar]]
                    if abs(zconc[-1][i] / stch_var[i, j]
                           ) < 10 and prop_flux[j] > 0:
                        lcri.add(j)
                    else:
                        lncr.add(j)
            else:
                lcri = set()

            # step 2
            epsilon = 0.03
            lcri = np.array(list(lcri))
            lncr = np.array(list(lncr))
            if not lncr:  # len(lncr) == 0:  # No non critical reactions
                dt1 = 1.0e+10
            else:
                alp = np.sum(prop_flux[lncr])
                uuj = np.matmul(stch_var[:, lncr], prop_flux[lncr])
                sig = np.matmul(stch_var2[:, lncr], prop_flux[lncr])
                exigi = np.array([cvar[j] * (1 / g_i[j](cvar[j]))
                                  for j in range(len(spc))]) \
                    * epsilon * del_coef
                exigi1 = np.maximum(exigi, np.full(len(spc), 1))
                dt1 = min(np.min(exigi1 / np.abs(uuj)),
                          np.min(exigi1 * exigi1 / np.abs(sig)))

            all_pos = False
            while not all_pos:
                kmul, d_t, do_ssa, _ = step_3to5(
                    prop_flux, lcri, dt1)  # step_3to5
                d_t = min(dto, d_t)
                if do_ssa:
                    t_c, tindex, index, break_now, all_pos, cvar = ssa_support(
                        tvar, ks_dict, r_dict, p_dict, stch_var, rfile,
                        tindex, index, t_c, zconc, spc, spc2, concz,
                        yconc, update_sp)
                    if break_now:
                        break
                else:
                    g_update = False
                    if index != tchlen:
                        if t_c + d_t > globals2.TCHECK[index]:
                            d_t = globals2.TCHECK[index] - t_c
                            index = index + 1
                            g_update = True

                    if t_c + d_t > tvar[tindex]:
                        d_t = max(0, tvar[tindex] - t_c)
                        kvar = np.round(
                            np.random.poisson(
                                prop_flux * d_t)) * kmul
                        all_pos = True
                        c_c = {}
                        b_b = np.sum(kvar * stch_var[:, update_sp], 0)
                        for xvar, _ in enumerate(spc):
                            holder = cvar[xvar] + b_b[xvar]
                            if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                                c_c[spc[xvar]] = holder
                            else:
                                all_pos = False
                                dt1 = dt1 / 2
                                break
                        if all_pos:
                            for xvar, _ in enumerate(spc):
                                concz[spc[xvar]] = c_c[spc[xvar]]
                            for xvar, _ in enumerate(spc2):
                                concz[spc2[xvar]] = concz[spc2[xvar]] + d_t
                            apply_rules(concz, yconc)
                            zconc.append([concz[xvar] for xvar in spc])
                            cvar = zconc[-1]
                            t_c = t_c + d_t
                            tindex = tindex + 1
                        else:
                            if g_update:
                                # pass
                                index = index - 1
                    else:
                        d_t = max(0, d_t)
                        kvar = np.round(
                            np.random.poisson(
                                prop_flux * d_t)) * kmul
                        all_pos = True
                        c_c = {}
                        b_b = np.sum(kvar * stch_var[:, update_sp], 0)
                        for xvar, _ in enumerate(spc):
                            holder = cvar[xvar] + b_b[xvar]
                            if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                                c_c[spc[xvar]] = holder
                            else:
                                all_pos = False
                                dt1 = dt1 / 2
                                break
                        if all_pos:
                            for xvar, _ in enumerate(spc):
                                concz[spc[xvar]] = c_c[spc[xvar]]
                            for xvar, _ in enumerate(spc2):
                                concz[spc2[xvar]] = concz[spc2[xvar]] + d_t
                            apply_rules(concz, yconc)
                            cvar = [concz[xvar] for xvar in spc]
                            t_c = t_c + d_t
                        else:
                            if g_update:
                                index = index - 1
                                # pass
            if tindex >= len(tvar):
                break
        if len(tvar) != len(zconc):
            while len(tvar) != len(zconc):
                zconc.append(zconc[-1])
        tnew = tvar
    return (tnew, np.array(zconc))


def step_3to5(prop_flux, lcri, dt1):
    """Additional steps in tau-leaping2"""
    # step 3
    kmul = np.ones((len(prop_flux), 1))
    r_1 = np.random.uniform()
    while r_1 == 0:
        r_1 = np.random.uniform()

    alpo = np.sum(prop_flux)
    do_ssa = False
    if dt1 < 10 * (1 / alpo):
        d_t = (1 / alpo) * (np.log(1 / r_1))
        do_ssa = True
    else:
        # step 4
        if lcri:  # > 0:
            alpc = np.sum(prop_flux[lcri])
            dt2 = (1 / alpc) * (np.log(1 / r_1))
        else:
            dt2 = 1.0e+5
            #dt2 = dt1

        # step 5
        if dt1 < dt2:
            d_t = dt1
            if lcri:  # > 0:
                kmul[lcri] = 0
        else:
            d_t = dt2
            if lcri:  # > 0:
                kmul[lcri] = 0
                pvar = np.cumsum([d / alpc for d in prop_flux[lcri]])
                r_2 = np.random.uniform()
                while r_2 == 0:
                    r_2 = np.random.uniform()
                for i, _ in enumerate(pvar):
                    if r_2 <= pvar[i]:
                        j_c = i
                        kmul[lcri[j_c]] = 1
                        break
    if d_t == 1.0e+5:
        do_ssa = True
    return [kmul, d_t, do_ssa, alpo]


def ssa_support(tvar, ks_dict, r_dict, p_dict, stch_var,
                rfile="", tindex=0, index=0, t_c=0, z_conc=[],
                spc=None, spc2=None, concz=None, yconc=None, update_sp=None):
    """SSA steps of tau-leaping2"""
    get_globals(rfile)
    stch_var = stch_var.T
    zconc = [[concz[z] for z in spc]]
    tmax = tvar[-1]

    break_now = False
    tchlen = len(globals2.TCHECK)
    all_pos = True
    for _ in range(100):
        if t_c < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            alp = np.sum(prop_flux)
            r_1 = np.random.uniform()
            while r_1 == 0:
                r_1 = np.random.uniform()
            pvar = np.cumsum([d / alp for d in prop_flux])
            d_t = (1 / alp) * (np.log(1 / r_1))

            #g_update = False
            if index != tchlen:
                if t_c + d_t >= globals2.TCHECK[index]:
                    d_t = globals2.TCHECK[index] - t_c
                    index = index + 1
                    #g_update = True

            if np.isnan(d_t) or np.isinf(d_t):
                z_conc.append(zconc[-1])
                while tvar[tindex] != tmax:
                    z_conc.append(zconc[-1])
                    tindex = tindex + 1
                t_c = tvar[-1]
                break_now = True
                break

            r_2 = np.random.uniform()
            for i, _ in enumerate(pvar):
                if r_2 <= pvar[i]:
                    all_pos = True
                    for xvar, _ in enumerate(spc):
                        holder = zconc[-1][xvar] + stch_var[i][update_sp[xvar]]
                        if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                            concz[spc[xvar]] = holder
                        else:
                            all_pos = False
                            break
                    if all_pos:
                        for xvar, _ in enumerate(spc2):
                            concz[spc2[xvar]] = concz[spc2[xvar]] + d_t
                        apply_rules(concz, yconc)
                        zconc.append([concz[xvar] for xvar in spc])
                        t_c = t_c + d_t
                        try:
                            if t_c == tvar[tindex]:
                                z_conc.append(zconc[-1])
                                tindex = tindex + 1
                            else:
                                while t_c > tvar[tindex]:
                                    z_conc.append(zconc[-2])
                                    tindex = tindex + 1
                        except:
                            pass
                    else:
                        for xvar, _ in enumerate(spc):
                            concz[spc[xvar]] = zconc[-1][xvar]
                        # if g_update:
                            #index = index - 1
                    break
        else:
            break_now = True
            break
    return [t_c, tindex, index, break_now, all_pos, zconc[-1]]



"""

                    This is the euler_mod module

The main  task  this module  do is to  integrate  ordinary  differential
equation (ODE)  using the Euler method. In  this module, two implementa-
tion is provided. Both are tau-adaptive.

The list of functions in this module are as follows;

1. euler_model
2. euler_int
3. euler2_model
4. euler_wer_est
5. euler_help
6. euler2_int

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.propagation.propensity import propensity_vec, \
    propensity_vec_molar
from BioSANS2020.propagation.recalculate_globals import get_globals, \
    apply_rules
from BioSANS2020.propagation.deterministic.lna_approx import \
    lna_steady_state
from BioSANS2020.myglobal import mglobals as globals2


def euler_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                stch_var, d_time, del_coef, molar=False):
    """This function returns the ODE model as a list of evaluated deri-
    vative of each components at a particular intance given by the state
    of the inputs.

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            stch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        d_time ([type]): [description]
        del_coef (float, optional): factor for modifying time steps used
            in the integration/propagation of ODE. Defaults to 10.
        molar : True if concentration or amount is in molar otherwise it
            is  False. Defaults to  False. If molar is True, macroscopic
            propensity is used. If it is False, microscopic propensity
            is used.

    Returns:
        list: 2 element list i.e. [value of dx/dt, d_time] where x are
            the components
    """

    if not molar:
        prop_flux = propensity_vec(ks_dict, conc, r_dict, p_dict, True)
    else:
        prop_flux = propensity_vec_molar(ks_dict, conc, r_dict, p_dict, True)

    fofx = np.matmul(stch_var, prop_flux)
    h1_div = np.abs(fofx) + 1.0e-30
    d_time = min(del_coef * np.min(1 / h1_div), d_time)
    dxdt = fofx * d_time
    if globals2.CON_BOUNDARY:
        spc = list(sp_comp.keys())  # [xvar for xvar in sp_comp]
        for xvar in globals2.CON_BOUNDARY:
            ind = spc.index(xvar)
            dxdt[ind] = 0
    return [dxdt.reshape(len(sp_comp)), d_time]


def euler_int(
        t_var, sp_comp, ks_dict, sconc, r_dict, p_dict, stch_var,
        del_coef=10, lna_solve=False, items=None, implicit=False,
        molar=False, rfile=""):
    """This function performs tau-adpative euler integration. The tau is
    adjusted in such a way that limit the change of the fastest reaction
    to del_coef amounts.

    Args:
        t_var (list): list of time points in the simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        sconc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stoch_var (numpy.ndarray): stoichiometric matrix. For example

            stoch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        del_coef (float, optional): factor for modifying time steps used
            in the integration/propagation of ODE. Defaults to 10.
        lna_solve (bool, optional): if True, proceed to linar noise
            approximation calculations. Defaults to False.
        items (list, optional): 3 item list of [canvas, scroll_x,
            scroll_y], Defaults to None.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        molar (bool, optional): True if concentration or amount is in
            molar and you want to use macroscopic equations otherwise it
            should be False and using microscopic equations. Defaults to
            False.
        rfile (str): file name of BioSANS topology file.

    Returns:
        tuple: (time, trajectory)
    """
    get_globals(rfile)
    tnew = []
    d_time = t_var[-1] - t_var[-2]
    yconc = {xvar: sconc[xvar] for xvar in sconc}
    conc = {xvar: sconc[xvar] for xvar in sconc}
    apply_rules(conc, yconc)
    straj_list = [[conc[z] for z in sp_comp]]

    if not implicit:
        tnow = t_var[0]
        tnew.append(tnow)
        while tnow < t_var[-1]:
            mvar = np.nan_to_num(euler_model(
                sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                d_time, del_coef, molar))
            mm_var = mvar[0].reshape(1, len(mvar[0]))[0]
            tnow = tnow + mvar[1]
            ind = 0
            for spi in sp_comp:
                conc[spi] = straj_list[-1][ind] + mm_var[ind]
                ind = ind + 1
            apply_rules(conc, yconc)
            straj_list.append([conc[z] for z in sp_comp])
            tnew.append(tnow)
        if lna_solve:
            return lna_steady_state(
                t_var, sp_comp, ks_dict, conc, r_dict, p_dict,
                stch_var, items=items)
    else:
        tnow = t_var[0]
        tnew.append(tnow)
        tindex = 1
        conc_list = [conc[z] for z in sp_comp]
        while tnew[-1] < t_var[-1]:
            mvar = np.nan_to_num(euler_model(
                sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                d_time, del_coef, molar))
            tnow = tnow + mvar[1]
            if tnow > t_var[tindex]:
                tnow = tnow - mvar[1]
                dt2 = t_var[tindex] - tnow
                mvar = np.nan_to_num(euler_model(
                    sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                    dt2, del_coef, molar))
                mm_var = mvar[0].reshape(1, len(mvar[0]))[0]
                tnow = tnow + mvar[1]

                ind = 0
                for spi in sp_comp:
                    conc[spi] = conc_list[ind] + mm_var[ind]
                    ind = ind + 1
                apply_rules(conc, yconc)
                conc_list = [conc[z] for z in sp_comp]
                for spi in sp_comp:
                    conc[spi] = max(0, conc[spi])
                straj_list.append(conc_list)

                tnew.append(tnow)
                tindex = tindex + 1
            else:
                mm_var = mvar[0].reshape(1, len(mvar[0]))[0]
                ind = 0
                for spi in sp_comp:
                    conc[spi] = conc_list[ind] + mm_var[ind]
                    ind = ind + 1
                apply_rules(conc, yconc)
                conc_list = [conc[z] for z in sp_comp]
                for spi in sp_comp:
                    conc[spi] = max(0, conc[spi])
    return (tnew, np.array(straj_list))


def euler2_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                 stch_var, molar=False):
    """[summary]

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stoch_var (numpy.ndarray): stoichiometric matrix. For example

            stoch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        molar (bool, optional): True if concentration or amount is in
            molar and you want to use macroscopic equations otherwise it
            should be False and using microscopic equations. Defaults to
            False.

    Returns:
        np.ndarray: derivative of components with respect to time at a
        particular time or instant based on the current state of the
        inputs.
    """
    if not molar:
        prop_flux = propensity_vec(ks_dict, conc, r_dict, p_dict, True)
    else:
        prop_flux = propensity_vec_molar(ks_dict, conc, r_dict, p_dict, True)

    dxdt = np.matmul(stch_var, prop_flux)
    if globals2.CON_BOUNDARY:
        spc = list(sp_comp.keys())  # [xvar for xvar in sp_comp]
        for xvar in globals2.CON_BOUNDARY:
            ind = spc.index(xvar)
            dxdt[ind] = 0
    return dxdt.reshape(len(sp_comp))


def euler_wer_est(h_var, sp_comp, ks_dict, conc, r_dict, p_dict,
                  stch_var, molar):
    """This function serves to do some part of the task in euler_help
    and euler2_int.
    """
    c_delta = h_var * euler2_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                                   stch_var, molar)
    yconc = {}
    ind = 0
    for xvar in sp_comp:
        yconc[xvar] = conc[xvar] + c_delta[ind]
        ind = ind + 1
    err = 0.5 * h_var \
        * euler2_model(sp_comp, ks_dict, yconc, r_dict, p_dict,
                       stch_var, molar) - 0.5 * c_delta
    return [yconc, err]


def euler_help(htry, eps, yscal, sp_comp, ks_dict, conc, r_dict, p_dict,
               stch_var, molar):
    """This function serves to do some part of the task in euler2_int.
    """
    safety = 0.9
    pgrow = -0.2
    pshrnk = -0.25
    errcon = 1.89e-4
    errmax = 1000
    h_var = htry
    while errmax > 1:
        ytemp, yerr = euler_wer_est(
            h_var, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var, molar)
        errmax = max(0, np.max(yerr / yscal)) / eps
        if errmax > 1:
            h_var = safety * h_var * (errmax**pshrnk)
            if h_var < 0.1 * h_var:
                h_var = 0.1 * h_var
        else:
            if errmax > errcon:
                hnext = safety * h_var * (errmax**pgrow)
            else:
                hnext = 5.0 * h_var
            hdid = h_var
            yvar = ytemp

    return [yvar, hdid, hnext, yerr]


def euler2_int(
        t_var, sp_comp, ks_dict, conc, r_dict, p_dict,
        stch_var, yscal=10, lna_solve=False, items=None, implicit=False,
        molar=False, rfile=""):
    """This function performs tau-adpative euler integration. The tau is
    adjusted to limit the error in each integration step as compared to
    what a second order runge-kutta would have predicted.

    Args:
        t_var (list): list of time points in the simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        sconc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stoch_var (numpy.ndarray): stoichiometric matrix. For example

            stoch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        yscal (float, optional): factor for modifying time steps used
            in the integration/propagation of ODE. Defaults to 10.
        lna_solve (bool, optional): if True, proceed to linar noise
            approximation calculations. Defaults to False.
        items (list, optional): 3 item list of [canvas, scroll_x,
            scroll_y], Defaults to None.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        molar (bool, optional): True if concentration or amount is in
            molar and you want to use macroscopic equations otherwise it
            should be False and using microscopic equations. Defaults to
            False.
        rfile (str): file name of BioSANS topology file.

    Returns:
        tuple: (time, trajectory)
    """

    get_globals(rfile)
    tnew = []
    delt = t_var[-1] - t_var[-2]
    straj_list = [[conc[z] for z in sp_comp]]
    eps = 1.0e-4

    if not implicit:
        tnow = t_var[0]
        tnew.append(tnow)
        tindex = 1
        while tnew[-1] < t_var[-1]:
            conc_old = {xvar: conc[xvar] for xvar in conc}
            conc, d_time, delt, _ = euler_help(
                delt, eps, yscal, sp_comp, ks_dict, conc, r_dict, p_dict,
                stch_var, molar)
            tnow = tnow + d_time
            if tnow > t_var[tindex]:
                tnow = tnow - d_time
                d_time = t_var[tindex] - tnow
                conc, _ = euler_wer_est(
                    d_time, sp_comp, ks_dict, conc_old, r_dict, p_dict,
                    stch_var, molar)
                delt = 5 * d_time
                tindex = tindex + 1
                tnow = tnow + d_time
            straj_list.append([conc[z] for z in sp_comp])
            tnew.append(tnow)

        if lna_solve:
            return lna_steady_state(
                t_var, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                items=items)
    else:
        tnow = t_var[0]
        tnew.append(tnow)
        tindex = 1
        while tnew[-1] < t_var[-1]:
            conc_old = {xvar: conc[xvar] for xvar in conc}
            conc, d_time, delt, _ = euler_help(
                delt, eps, yscal, sp_comp, ks_dict,
                conc, r_dict, p_dict, stch_var, molar)
            tnow = tnow + d_time
            if tnow > t_var[tindex]:
                tnow = tnow - d_time
                d_time = t_var[tindex] - tnow
                conc, _ = euler_wer_est(
                    d_time, sp_comp, ks_dict, conc_old, r_dict, p_dict,
                    stch_var, molar)
                delt = 5 * d_time
                straj_list.append([conc[z] for z in sp_comp])
                tnow = tnow + d_time
                tnew.append(tnow)
                tindex = tindex + 1
    return (tnew, np.array(straj_list))



"""

                     This is the ode_extract module

This module grabs stoichiometric matrix, propensity vector, and chemical
reaction from a list of strings containing python expression

List of functions:

1. process
2. prop_extr
3. term_ext
4. get_prop_stoich
5. print_stoich_prop
6. grab_rxn_ksn
7. transform_to_rxn
8. odedxdt_to_topo

For a given set of chemical species and corresponding ODE expression,
this script can be used as follows;

1.	Declare the list of ODE and variables or grab the list from a file.

    x = ["m","P"]
    dx_dt = ["km - m*rm","kp*m - P*rp"]

2.	Use "transform_to_rxn" command to get reactions that match the ODE.

    transform_to_rxn(x, dx_dt)

The result of this command is as follows;

    0 NONE => 1.0 m, 1 ::::: lambda   : km
    1.0 m => 0 NONE, 1 ::::: lambda m : m*rm
    1.0 P => 0 NONE, 1 ::::: lambda P : P*rp
    0 NONE => 1.0 P, 1 ::::: lambda m : kp*m

The first line corresponds to the formation of mRNA m with a rate
constant of km. The second line is degradation of m with degradation
constant rm. The third line is degradation of protein with degradation
constant rp. The last line is the formation of P with rate constant kp.
The lambda expression after the delimiter ':::::' tells us the
propensity for each reaction which is provided after ':'. The 0 NONE
means there is no defined reactant and or degradation product for this
reaction.

3.	Use "print_stoich_prop" to print stoichiometric matrix and
    propensity vector.

        print_stoich_prop(dx_dt)

    The result is as follows;

    1.0*km - 1.0*m*rm
    -1.0*P*rp + 1.0*kp*m

    [1.00000000000000, -1.00000000000000, 0, 0]
    [0, 0, -1.00000000000000, 1.00000000000000]

    [km]
    [m*rm]
    [P*rp]
    [kp*m]

The first two line here is the declared ODE. After the space, is the
stoichiometric matrix. The rows corresponds to the species declared in
step 1 i.e. first row in the matrix corresponds to x[0] or m and the
second row corresponds to x[1] or P. The columns correspond to the terms
in the propensity as listed after the stoichiometric matrix.


"""

from sympy import sympify, expand, Matrix
import numpy as np
from BioSANS2020.gui_functs.scrollable_text import INSERT
from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text


DONE_PARSING = set()


def process(xvar):
    """This function formats the expression xvar into a form that can be
    easily processed by stripping and replacing invalid operators.
    Args:
        xvar : mathematical expresson i.e. 'A*B*ka/(C**2 + 1)'
    Returns:
        val : same as xvar but with invalid operator removed
    """
    # global DONE_PARSING

    # stripping leading and trailing * operator
    # replacing */ with /
    xxvar = xvar.strip("*").strip().replace("*/", "/")
    if xxvar[0] == "/":
        xxvar = "1" + xxvar
    if xxvar[-1] == "/":
        xxvar = xxvar.strip("/")
    val = sympify(xxvar.strip("*"))
    if val in DONE_PARSING:
        return None
    # else:
    DONE_PARSING.add(val)
    return val


def prop_extr(expr, prop):
    """This function extracts the propensity terms from the expression
    expr which is a string of python mathematical expression. Propensity
    terms are the terms that can be separated by + or - without the
    numerical coefficient or stoichiometric multiplier.
    Args:
        expr : mathematical expresson i.e. 'A*B*ka/(C**2 + 1)'
        prop : propensity list to append extracted propensity term
    """
    ex = str(expr)  # transform to string to ensure it is string
    open_p = 0      # number of opening parenthesis
    close_p = 0     # number of closing parenthesis
    diff = 0        # difference between open_p and close_p

    collect = ""    # variable to collect propensity terms
    # iteration in each character of the string exprression ex
    for vstr in ex:
        if vstr in ["+", "-"] and diff == 0:
            if len(collect) > 0:
                dvar = process(collect)
                if dvar is not None:
                    prop.append(dvar)
                collect = ""
                open_p = 0
                close_p = 0
        elif vstr == "(":
            open_p = open_p + 1
            collect = collect + vstr
        elif vstr == ")":
            close_p = close_p + 1
            collect = collect + vstr
        elif (vstr.isnumeric() or vstr == ".") and (diff == 0):
            if len(collect) >= 3:
                if collect[-2:] == "**" or collect[-1] not in ["*", "/"]:
                    collect = collect + vstr
            elif len(collect) >= 1:
                if collect[-1] not in ["*", "/", " "]:
                    collect = collect + vstr
            else:
                pass
        else:
            collect = collect + vstr
        diff = open_p - close_p

    dvar = process(collect)
    if dvar is not None:
        prop.append(dvar)


def term_ext(expr):
    """This function extracts the terms from the expression expr which
    is a string of python mathematical expression. The terms are part of
    expr which can be separated by + or - with the numerical coefficient
    or stoichiometric multiplier.
    Args:
        expr : mathematical expresson i.e. 'A*B*ka/(C**2 + 1)'
    Returns:
        term : the list of extracted terms
    """
    term = []
    ex = str(expr)
    open_p = 0
    close_p = 0
    diff = 0

    collect = ""
    last_sign = ""
    for vstr in ex:
        if vstr in ["+", "-"] and diff == 0:
            if len(collect) > 0:
                term.append(last_sign + collect)
                collect = ""
                open_p = 0
                close_p = 0
            last_sign = vstr
        elif vstr == "(":
            open_p = open_p + 1
            collect = collect + vstr
        elif vstr == ")":
            close_p = close_p + 1
            collect = collect + vstr
        else:
            collect = collect + vstr
        diff = open_p - close_p

    term.append(last_sign + collect)
    return term


def get_prop_stoich(dxdt):
    """This function extracts the propensity vector and stoichiometric
    matrix from the list of ordinary differential equation.
    Args:
        dxdt : list of strings of mathematical expression
    Returns :
        v_stoich : stoichiometric matrix (sympy Matrix)
        w_var : propensity vector ( sympy type Matrix )
    """
    prop = []
    da_dt = []
    for expro in dxdt:
        expr = expand(sympify(expro))
        prop_extr(expr, prop)
        da_dt.append(expr)

    w_var = prop
    v_stoich = [[0 for xvar in range(len(w_var))] for y in range(len(da_dt))]

    for i, _ in enumerate(da_dt):
        for j, _ in enumerate(w_var):
            s_var = 0
            for xxvar in term_ext(da_dt[i]):
                xvar = sympify(xxvar) / w_var[j]
                try:
                    s_var = s_var + float(xvar)
                except:
                    pass
            v_stoich[i][j] = s_var

    return Matrix(v_stoich), Matrix(w_var)


def print_stoich_prop(dxdt):
    """This function extracts the propensity vector and stoichiometric
    matrix from the list of ordinary differential equation and prints
    the output in the console.
    Args:
        dxdt : list of strings of mathematical expression
    """
    global DONE_PARSING
    DONE_PARSING = set()
    print()
    v_stoich, w_var = get_prop_stoich(dxdt)
    for tvar in v_stoich * w_var:
        print(tvar)
    print()

    for c_var in np.array(v_stoich):
        print([round(y, 4) for y in c_var])
    print()

    for c_var in np.array(w_var):
        print(c_var)


def grab_rxn_ksn(stch_var, xvar, w_var):
    """This function transform the stoichiometric matrix, species or
    components, and propensity vector into a list of reactions and list
    of rate constants.
    Args:
        stch_var : stoichiometric matrix or 2D Matrix of coefficient
        xvar     : list of components or species
        w_var    : propensity vector or 1D Matrix of fluxes
    Returns:
        rxn_var  : list of reaction similar to BioSANS reactions format
        ksn_var  : set of rate constant symbols if not numeric
    """
    rxn_var = []
    ksn_var = set()
    ind = 0

    for col in stch_var.T:
        r_big = ""
        p_big = ""
        for i, _ in enumerate(col):
            if col[i] != 0:
                if col[i] < 0:
                    r_big += str(abs(col[i])) + " " + xvar[i] + " " + "+ "
                else:
                    p_big += str(abs(col[i])) + " " + xvar[i] + " " + "+ "
        if r_big.strip() == "":
            r_big = "0 NONE"
        if p_big.strip() == "":
            p_big = "0 NONE"

        in_sp = []
        for s_var in w_var[ind].free_symbols:
            ss_var = str(s_var)
            if ss_var in xvar:
                in_sp.append(ss_var)
            else:
                ksn_var.add(ss_var)
        in_sp = ",".join(in_sp)
        rxn_var.append(r_big.strip("+ ") + " => " + p_big.strip("+ ") +
                       ", 1 ::::: lambda " + in_sp + " : " + str(w_var[ind]))
        ind = ind + 1

    return rxn_var, ksn_var


def transform_to_rxn(xvar, dxdt, x_ini=None, k_rc=None, items=None):
    """This function transform the list of components and list of ODE
    into BioSANS topology file format.
    Args:
        xvar     : list of components or species (variable in ODE)
                   example ["m","P"]
        dxdt     : list of differential equations (python string)
                   example ["km - m*rm","kp*m - P*rp"]
        x_ini    : dictionary of initial conecentration or value
                   example {"m" : 10, "p" : 0}
        k_rc     : dictionary of rate constant symbols abd values
                   example {"km" : 0.1, "rm" : 0.2}
        items    : list containing [canvas, scroll_x, scroll_y]
    Returns:
        text     : text area where the outputs are written
    """
    global DONE_PARSING
    # print(xvar, dxdt, x_ini, k_rc, items, sep="\n\n")
    if items:
        text = prepare_scroll_text(items)

        def ffrint(xvar):
            return text.insert(INSERT, xvar + "\n")
    else:
        def ffrint(xvar):
            return print(" ".join([str(y) for y in xvar]), end="")

    if x_ini is None:
        x_ini = {}
    if k_rc is None:
        k_rc = {}

    DONE_PARSING = set()
    v_stoich, w_var = get_prop_stoich(dxdt)
    stch_var = np.around(np.array(v_stoich).astype(float), 3)
    rxn_var, ksn_var = grab_rxn_ksn(stch_var, xvar, w_var)

    ffrint("Function_Definitions:")
    for k in ksn_var:
        if k.strip() not in k_rc:
            ffrint(k.strip() + " = type actual value")
        else:
            ffrint(k.strip() + " = " + k_rc[k.strip()])
    for c_var in xvar:
        if c_var.strip() not in x_ini:
            ffrint(c_var.strip() + "_ini = type actual value")
        else:
            ffrint(c_var.strip() + "_ini = " + x_ini[c_var.strip()])

    ffrint("")
    ffrint("#REACTIONS")
    for rx_i in rxn_var:
        ffrint(rx_i)

    ffrint("")
    ffrint("@CONCENTRATION")
    for c_var in xvar:
        ffrint(c_var + " , " + c_var.strip() + "_ini")
    ffrint("NONE, 1")
    return text


def odedxdt_to_topo(mfile, items):
    """This function reads BioSANS ODE file format and converts it to
    BioSANS topology file format. The ODE file contains ODE expression,
    initial concentration, and rate constants separated by tags. There
    are three tags that is currently supported : ODE_DECLARATIONS,
    INI_CONCENTRATIONS, and RATE_CONSTANTS. Provided below is an example
    of the content of a typical ODE file.

        ODE_DECLARATIONS:
        A = -ka*A*B/(1+C**2) + kf1/(1+B**2)
        B = -ka*A*B/(1+C**2)
        C = -kc*C + kf2
        D = ka*A*B/(1+C**2) - kf2

        INI_CONCENTRATIONS:
        A = 100
        B = 200
        C = 150
        D = 0

        RATE_CONSTANTS:
        ka = 0.02
        kf1 = 0.2
        kc = 0.03
        kf2 = 0.01

    Args:
        mfile    : file containing list of ODE expression
        items    : list containing [canvas, scroll_x, scroll_y]
    Returns:
        text     : text area where the outputs are written
    """
    ddvar = open(mfile, "r")
    print()
    xvar = []
    x_ini = {}
    k_rc = {}
    dxdt = []
    last = ""
    for xxvar in ddvar:
        if last == "ODE_DECLARATIONS" and xxvar.strip(
        ) not in ["INI_CONCENTRATIONS:", "RATE_CONSTANTS:"]:
            if xxvar.strip() != "":
                row = xxvar.split("=")
                xvar.append(row[0].strip())
                dxdt.append(row[1])
        elif last == "INI_CONCENTRATIONS" and xxvar.strip() \
                not in ["ODE_DECLARATIONS:", "RATE_CONSTANTS:"]:
            if xxvar.strip() != "":
                row = xxvar.split("=")
                x_ini[row[0].strip()] = row[1].strip()
        elif last == "RATE_CONSTANTS" and xxvar.strip() \
                not in ["ODE_DECLARATIONS:", "INI_CONCENTRATIONS:"]:
            if xxvar.strip() != "":
                row = xxvar.split("=")
                k_rc[row[0].strip()] = row[1].strip()
        elif xxvar.strip() == "ODE_DECLARATIONS:":
            last = "ODE_DECLARATIONS"
        elif xxvar.strip() == "INI_CONCENTRATIONS:":
            last = "INI_CONCENTRATIONS"
        elif xxvar.strip() == "RATE_CONSTANTS:":
            last = "RATE_CONSTANTS"

    # print_stoich_prop(dxdt)
    ddvar.close()
    return transform_to_rxn(xvar, dxdt, x_ini, k_rc, items)

# odedxdt_to_topo("NewText.txt")



# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

"""
                       This is the transform_data module

This module process numeric trajectories and either prints the result or
display the output into a text area embedded in a canvas or plots it

The following is the list of function for this module:

1. normalize
2. calc_cross_corr
3. calc_covariance2
4. calc_covariance
5. fano_factor
6. prob_density_calc
7. prob_density_calc2
8. prob_density_calc3
9. ave_traj_calc

"""

import matplotlib.pyplot as plt
import numpy as np
from BioSANS2020.analysis.numeric.sample_points import sample_points
from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.gui_functs.scrollable_text import INSERT
from BioSANS2020.gui_functs.draw_figure import draw_figure
# from mpl_toolkits.mplot3d import Axes3D
# from scipy.stats import norm
from BioSANS2020.myglobal import mglobals as globals2


def normalize(vect):
    """ returns the normalized form of the input vector v """
    norm = np.linalg.norm(vect)
    if norm == 0:
        return vect
    return vect / norm


def calc_cross_corr(edata, items):
    """This function calculates the cross correlation of edata[0] and
    returns a plot of the correlation as a function of lags.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
    """
    if len(edata[0]) != 3:
        ndata, slabels = sample_points(edata)
        # time = ndata[0]
        data_length = len(ndata)
        dlen = len(slabels)

        ddm = [0] * len(slabels)
        for ind in range(1, data_length):
            ith_data = ndata[ind]
            for jnd in range(dlen):
                ddm[jnd] = ith_data[jnd] + ddm[jnd]

        ddm = [ddm[jnd] / (data_length - 1) for jnd in range(dlen)]
        vvv = [[[0] for xvar in range(dlen)] for xvar in range(dlen)]
        for ind in range(1, data_length):
            ith_data = ndata[ind]
            for jnd in range(dlen):
                for knd in range(dlen):
                    vvv[jnd][knd] = vvv[jnd][knd] + np.correlate(
                        normalize(ith_data[jnd][-1000:] - ddm[jnd][-1000:]),
                        normalize(ith_data[knd][-1000:] - ddm[knd][-1000:]),
                        "full",
                    )
        vvv = np.array(vvv) / (data_length - 1)
        for jnd in range(dlen):
            for knd in range(dlen):
                plt.figure(figsize=(9.68, 3.95))
                plt.xlabel("lag")
                plt.ylabel(slabels[jnd] + "-" + slabels[knd] + " correlation")
                xvar = int(len(vvv[jnd, knd]) / 2)
                xvar = [-y for y in list(range(xvar))][::-1] \
                    + [0] + list(range(xvar))
                line = plt.plot(xvar, vvv[jnd, knd])
                fig = plt.gcf()
                plt.tight_layout()
                globals2.PLOTTED.append([plt.gca(), fig, [line]])
                draw_figure(items, fig)
                # plt.close()


def calc_covariance2(edata):
    """This function calculates the covariance of edata[0] and prints
    the result in a terminal window.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
    """
    ndata, slabels = sample_points(edata)
    # time = ndata[0]
    data_length = len(ndata)
    dlen = len(slabels)

    ddm = [0] * len(slabels)
    for ind in range(1, data_length):
        data_slice = np.array(ndata[ind])[:, -100:]
        for jnd in range(dlen):
            ddm[jnd] = data_slice[jnd] + ddm[jnd]

    ddm = [ddm[jnd] / (data_length - 1) for jnd in range(dlen)]
    ddm = np.mean(np.array(ddm), 1).reshape(dlen, 1)

    cumulative_cov = 0
    for ind in range(1, data_length):
        data_slice = np.array(ndata[ind])[:, -100:] - ddm
        ddd = np.zeros((dlen, dlen))
        for jnd in range(dlen):
            for knd in range(dlen):
                ddd[jnd, knd] = np.mean(data_slice[jnd] * data_slice[knd])
        cumulative_cov = cumulative_cov + ddd
    cumulative_cov = cumulative_cov / (data_length - 1)
    print("covariance")
    print(slabels)
    print(cumulative_cov)
    print()
    vvv = np.zeros((dlen, dlen))
    for ind in range(dlen):
        for jnd in range(dlen):
            vvv[ind, jnd] = cumulative_cov[ind, jnd] \
                / np.sqrt(cumulative_cov[ind, ind] * cumulative_cov[jnd, jnd])
    print("\ncorrelation")
    print(slabels)
    print(vvv)


def calc_covariance(edata, items, points=100):
    """This function calculates the covariance of edata[0] and prints
    the result in a text area embedded in a canvas.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
        points : last number of points considered in covariance
                 calculation from -points to the end of array
                 or equivalent to [-points:] slice.
    """
    data, slabels = edata
    data_length = len(data)
    ddm = 0
    for ind in range(1, data_length):
        data_slice = data[ind][-points:, 1:]
        ddm = ddm + data_slice

    ddm = np.mean(ddm / (data_length - 1), 0)
    # print(ddm)

    cumulative_cov = 0
    dlen = len(slabels)
    for knd in range(1, data_length):
        data_slice = data[knd][-points:, 1:] - ddm
        ddd = np.zeros((dlen, dlen))
        for ind in range(dlen):
            for jnd in range(dlen):
                ddd[ind, jnd] = np.mean(data_slice[:, ind]
                                        * data_slice[:, jnd])
        cumulative_cov = cumulative_cov + ddd
    cumulative_cov = cumulative_cov / (data_length - 1)

    text = prepare_scroll_text(items)
    text.insert(INSERT, "covariance\n\n")
    for ind in range(dlen):
        for jnd in range(ind, dlen):
            if str(cumulative_cov[ind, jnd]) not in {"None", "nan", "0.0"}:
                label = slabels[ind] + " and " + slabels[jnd]
                text.insert(INSERT, label.ljust(50) + " = "
                            + str(cumulative_cov[ind, jnd]) + "\n")

    vvv = np.zeros((dlen, dlen))
    text.insert(INSERT, "\ncorrelation\n\n")
    for ind in range(dlen):
        for jnd in range(ind, dlen):
            if str(cumulative_cov[ind, jnd]) not in {"None", "nan", "0.0"}:
                vvv[ind, jnd] = cumulative_cov[ind, jnd] \
                    / np.sqrt(cumulative_cov[ind, ind]
                              * cumulative_cov[jnd, jnd])
                label = slabels[ind] + " and " + slabels[jnd]
                text.insert(INSERT, label.ljust(50) + " = "
                            + str(vvv[ind, jnd]) + "\n")

    data_slice = 0
    for ind in range(data_length):
        data_slice = data_slice + (data[ind][-points:, 1:] - ddm) ** 2
    fano_f = np.mean(data_slice / (data_length - 1), 0) / ddm
    text.insert(INSERT, "\nFano Factor\n\n")
    for ind in range(dlen):
        if str(fano_f[ind]) not in {"None", "nan"}:
            text.insert(INSERT, slabels[ind].ljust(50) + " = "
                        + str(fano_f[ind]) + "\n")

    globals2.PLOT_I = globals2.PLOT_I + 1


def fano_factor(edata, items, points=100):
    """This function calculates the fano-factor of edata[0] and prints
    the result in a text area embedded in a canvas.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
        points : last number of points considered in fano-factor
         calculation from -points to the end of array ([-points:] slice)
    """
    data, slabels = edata
    data_length = len(data)
    ddm = 0
    for ind in range(data_length):
        data_slice = data[ind][-points:, 1:]
        ddm = ddm + data_slice

    ddm = np.mean(ddm / (data_length - 1), 0)

    data_slice = 0
    for ind in range(data_length):
        data_slice = data_slice + (data[ind][-points:, 1:] - ddm) ** 2
    fano_f = np.mean(data_slice / (data_length - 1), 0) / ddm

    text = prepare_scroll_text(items)
    dlen = len(slabels)
    text.insert(INSERT, "Fano Factor\n\n")
    for ind in range(dlen):
        if str(fano_f[ind]) not in {"None", "nan"}:
            text.insert(INSERT, slabels[ind].ljust(50) + " = "
                        + str(fano_f[ind]) + "\n")

    globals2.PLOT_I = globals2.PLOT_I + 1


def prob_density_calc(edata, items):
    """This function calculates the probability density of edata[0] and
    returns a plot of the probability density.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
    """
    data, slabels = edata
    data_length = len(data)
    dlen = len(slabels)

    # data_slice is a concatenation of all the rows of  trajectory data
    data_slice = data[0][:, 1:]
    for ind in range(1, data_length):
        data_slice = np.concatenate((data_slice, data[ind][:, 1:]), axis=0)

    for jnd in range(dlen):
        plt.figure(figsize=(9.68, 3.95))
        plt.xlabel("conc(" + slabels[jnd] + ")")
        plt.ylabel("Prob")
        line = plt.hist(data_slice[:, jnd], bins=100, density=True,
                        orientation="vertical")
        plt.legend([slabels[jnd]])
        plt.tight_layout()
        fig = plt.gcf()
        globals2.PLOTTED.append([plt.gca(), fig, [line]])
        draw_figure(items, fig)
        # plt.close()


def prob_density_calc2(edata, items):
    """This function calculates the probability density of edata[0] and
    returns a plot of the probability density with time.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
    """
    data, slabels = edata
    data_length = len(data)
    dlen = len(slabels)

    data_slice = data[0][:, 1:]
    data_list = data[0][:, 0].flatten()
    # tlen = len(data_list)
    for ind in range(1, data_length):
        data_slice = np.concatenate((data_slice, data[ind][:, 1:]), axis=0)
        data_list = np.concatenate((data_list, data[ind][:, 0]))

    for jnd in range(dlen):
        # r = data_slice
        zzs, xxs, yys = np.histogram2d(data_list, data_slice[:, jnd].flatten(),
                                       bins=(100, 100))
        plt.figure(figsize=(9.68, 3.95))
        plt.xlabel("time")
        plt.ylabel("conc(" + slabels[jnd] + ")")
        cntr = plt.contour(
            xxs[1:],
            yys[1:],
            zzs.T,
            list(range(int(np.min(zzs)), int(np.max(zzs)) + 1, 2)),
            cmap="Spectral_r",
        )
        fig = plt.gcf()
        fig.colorbar(cntr)
        plt.tight_layout()
        globals2.PLOTTED.append([plt.gca(), fig, [cntr]])
        draw_figure(items, fig)
        # plt.close()


def prob_density_calc3(edata, items, bins=50):
    """This function calculates the probability density of edata[0] per
    bins and returns a plot of the probability density (time slice).
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
        bins : number of bins an entire trajectory will be discretized
    """
    if len(edata[0]) != 3:
        ndata, slabels = sample_points(edata)
        time = ndata[0]
        data_length = len(ndata)
        dlen = len(slabels)

        ddd = [[] for xvar in range(dlen)]
        for knd in range(1, data_length):
            data_slice = ndata[knd]
            for jnd in range(dlen):
                ddd[jnd].append(data_slice[jnd])
        ddd = np.array(ddd)

        vvv = {}
        for ind in range(dlen):
            for jnd in range(len(time)):
                vvv[(ind, jnd)] = np.histogram(ddd[ind][:, jnd], bins=bins)

        tlen = len(time)
        strt = int(tlen / 10)
        for ind in range(dlen):
            lines = []
            fig = plt.figure(figsize=(9.68, 3.95))
            axf = fig.gca(projection="3d")
            for jnd in range(strt, tlen, int(tlen / 5)):
                xxs = vvv[(ind, jnd)][1][1:]
                yys = vvv[(ind, jnd)][0]
                width = xxs[1] - xxs[0]
                line = axf.bar(xxs, yys, zs=time[jnd], zdir="xvar",
                               width=max(width, 0.8), alpha=1)
                lines.append(line)
            axf.set_xlabel("time")
            axf.set_ylabel("conc(" + slabels[ind] + ")")
            axf.set_zlabel("freq")
            axf.view_init(elev=40, azim=-120)
            globals2.PLOTTED.append([axf, fig, lines])
            draw_figure(items, fig)
            # plt.close()


def ave_traj_calc(edata, items):
    """This function calculates the average trajectory of edata[0] and
    returns a plot of the average trajectory as a function of time.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
    """
    if len(edata[0]) != 3:
        ndata, slabels = sample_points(edata)
        time = ndata[0]
        data_length = len(ndata)
        dlen = len(slabels)

        ddm = [0] * len(slabels)
        for ind in range(1, data_length):
            data_slice = ndata[ind]
            for jnd in range(dlen):
                ddm[jnd] = data_slice[jnd] + ddm[jnd]

        ddm = [ddm[jnd] / (data_length - 1) for jnd in range(dlen)]

        plt.figure(figsize=(9.68, 3.95))
        plt.xlabel("time (sec)")
        plt.ylabel("conc")
        lines = []
        for jnd in range(dlen):
            line = plt.plot(time, ddm[jnd], label=slabels[jnd])
            lines.append(line)
        plt.legend()
        fig = plt.gcf()
        plt.tight_layout()
        globals2.PLOTTED.append([plt.gca(), fig, lines])
        draw_figure(items, fig)
        # plt.close()



"""

            This module is the create_wxmaxima_command module

This module attempts to transforms topology file into wxmaxima ODE input
that a user can copy and paste into wxmaxima to grab a symbolic solution


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from sympy import Function, Symbol, Matrix

from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.propagation.propensity import propensity_vec_molar
from BioSANS2020.propagation.recalculate_globals import get_globals
# from BioSANS2020.myglobal import mglobals as globals2


def prepare_ffrrint(items):
    """This function instantiate how the output will be printed by crea-
    thing the function ffprint

    Args:
        text (Text): text area where the outputs are written
        items (tuple): 3 item list of (canvas, scroll_x, scroll_y)

    Returns:
        function: either print in console or in text area
    """
    if items:
        text = prepare_scroll_text(items)

        def ffprint(xvar):
            return text.insert('insert', " ".join([str(y) for y in xvar]))
    else:

        def ffprint(xvar):
            return print(" ".join([str(y) for y in xvar]), end="")
    return ffprint


def prepare_dict1(sp_comp):
    """This function prepares cs_var, cso_var, time

    Args:
        sp_comp (dict): dictionary of components position in reaction

    Returns:
        multitype: cs_var, cso_var, time
    """
    cs_var = {}
    cso_var = {}
    time = Symbol('t', real=True, positive=True)
    for xvar in sp_comp:
        cs_var[xvar] = Function(xvar, positive=True)
        cs_var[xvar] = cs_var[xvar](time)
        cso_var[xvar] = Symbol(xvar + "o", real=True, negative=False)
    return cs_var, cso_var, time


def prepare_list1(ks_dict):
    """This function prepares a list of symbol for rate constant

    Args:
        ks_dict (dict): dictionary of rate constant positions

    Returns:
        list: rate constants symbols
    """
    kcs = []
    for i, _ in enumerate(ks_dict):
        row = []
        if len(ks_dict[i]) == 1:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
        else:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
            key = 'kb' + str(i + 1)
            row.append(Symbol(key, real=True, positive=True))
        kcs.append(row)
    return kcs


def prepare_list2(stch_var, slabels):
    """This function prepares two list from stoichiometric matrix

    Args:
        stch_var (np.ndarray): toichiometric matrix

    Returns:
        list: [non zero row values of  stch_var,
            nonzero rows index of stch_var]
    """
    ss_var = []
    nz_var = []
    for row in range(stch_var.shape[0]):
        if sum(abs(stch_var[row, :])) != 0 and slabels[row][0] != "-":
            ss_var.append(list(stch_var[row, :]))
            nz_var.append(row)
    ss_var = Matrix(ss_var)
    return ss_var, nz_var


def for_wxmaxima(sp_comp, ks_dict, conc, r_dict, p_dict,
                 stch_var, items=None, rfile=""):
    """This function  transforms topology file  into wxmaxima  ODE input
    that a user can copy and paste into wxmaxima to grab a symbolic
    solution.

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        items (list): list of [canvas, scroll_x, scroll_y]
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        list: [0, 0] - contains nothing
    """
    get_globals(rfile)
    ffprint = prepare_ffrrint(items)

    ffprint(["/*Copy and paste to wxmaxima and run the cell*/\n\n"])
    cs_var, cso_var, time = prepare_dict1(sp_comp)
    kcs = prepare_list1(ks_dict)

    prop_flux = Matrix(propensity_vec_molar(kcs, cs_var, r_dict, p_dict, True))
    stch_var = Matrix(stch_var)
    # cs_var might have change after call
    for xvar in sp_comp:
        cs_var[xvar] = Function(xvar, positive=True)
        cs_var[xvar] = cs_var[xvar](time)
    slabels = list(cs_var.keys())  # [xvar for xvar in cs_var]

    ss_var, nz_var = prepare_list2(stch_var, slabels)

    da_dt = ss_var * prop_flux
    ccs = [cs_var[xvar] for xvar in cs_var]
    ccso = [cso_var[xvar] for xvar in cso_var]
    js_var = [ccs[xvar] for xvar in nz_var]
    jso = [ccso[xvar] for xvar in nz_var]

    ds_dt2 = []
    for xvar in js_var:
        ds_dt2.append(xvar.diff(time))

    fe_var = []
    for i, _ in enumerate(ds_dt2):
        ffprint([
            "f" + str(i + 1) + ":" + str(ds_dt2[i])
            .replace("Derivative", "diff") + " = " + str(da_dt[i]) + ";",
            "\n"])

        fe_var.append("f" + str(i + 1))
    ffprint(["\n"])
    for xvar, _ in enumerate(js_var):
        ffprint([
            "atvalue(" + str(js_var[xvar]) + ",t=0," + str(jso[xvar]) + ");",
            "\n"])

    ffprint(["\ndesolve(" + str(fe_var).replace('"', '').replace("'", '') +
             "," + str(js_var) + ");"])
    ffprint(["\n"])
    ffprint(["\n\n/*Copy and paste to wxmaxima and run the cell*/\n"])
    ffprint(["\n"])

    for xvar in sp_comp:
        cso_var[xvar] = conc[xvar]

    prop_flux = Matrix(
        propensity_vec_molar(ks_dict, cs_var, r_dict, p_dict, True))
    stch_var = Matrix(stch_var)
    # cs_var might have change after call
    for xvar in sp_comp:
        cs_var[xvar] = Function(xvar, positive=True)
        cs_var[xvar] = cs_var[xvar](time)
    slabels = list(cs_var.keys())  # [xvar for xvar in cs_var]
    ss_var = []
    nz_var = []
    for row in range(stch_var.shape[0]):
        if sum(abs(stch_var[row, :])) != 0 and slabels[row][0] != "-":
            ss_var.append(list(stch_var[row, :]))
            nz_var.append(row)
    ss_var = Matrix(ss_var)

    da_dt = ss_var * prop_flux
    ccs = [cs_var[xvar] for xvar in cs_var]
    ccso = [cso_var[xvar] for xvar in cso_var]
    js_var = [ccs[xvar] for xvar in nz_var]
    jso = [ccso[xvar] for xvar in nz_var]

    ds_dt2 = []
    for xvar in js_var:
        ds_dt2.append(xvar.diff(time))

    fe_var = []
    for i, _ in enumerate(ds_dt2):
        ffprint([
            "f" + str(i + 1) + ":" + str(ds_dt2[i])
            .replace("Derivative", "diff") + " = " + str(da_dt[i]) + ";",
            "\n"])
        fe_var.append("f" + str(i + 1))
    ffprint(["\n"])
    for xvar, _ in enumerate(js_var):
        ffprint([
            "atvalue(" + str(js_var[xvar]) + ",t=0," + str(jso[xvar]) + ");",
            "\n"])

    ffprint(["\ndesolve(" + str(fe_var).replace('"', '').replace("'", '') +
             "," + str(js_var) + ");"])

    return [0, 0]



"""

                     This is the lna_approx module

     This module performs numerical linear noise approximation or LNA by
exploiting the following relationship.

                         AC + CA.T + BB = 0

where A is defined as d(S*f)/dx where S is the stoichiometric  matrix, f
are the propensities or fluxes, and x are the components or species. The
flux is a function of  x and rate constant k. BB is the diffusion matrix
equivalent  to S*diag(f)*S.T where diag(f) is a  square matrix with zero
non-diagonal elements and f[i] in each diagonal elements.

The following are the list of function for this module.

1. rem_rowcol_zero
2. lna_ss_jacobian
3. lna_model_ss
4. lna_steady_state


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from scipy import linalg as LA
from scipy.optimize import fsolve

from BioSANS2020.propagation.propensity import propensity_vec_molar
from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text


def rem_rowcol_zero(a_mat):
    """This function removes rows and columns without non-zero entries.

    Args:
        a_mat (np.ndarray): numpy matrix of A or d(S*f)/dx as described
            in the module docstring.

    Returns:
        np.ndarray: numpy matrix with no rows and columns without non-
            zero entries.
    """
    return a_mat[:, ~np.all(a_mat == 0, axis=0)][~np.all(a_mat == 0, axis=1)]


def lna_ss_jacobian(model, zlist, sp_comp, stch_var,
                    ks_dict, r_dict, p_dict):
    """This  function calculataes  the  jacobian  of  the model ODE with
    respect to the list of species concentration zlist.

    Args:
        model (function): the ODE model returning derivative of species
            or components as a function of time
        zlist (list): list of components or species amounts
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        stch_var (np.ndarray): stoichiometric matrix
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }

    Returns:
        np.ndarray: jacobian matrix or d(S*f)/dx or A
    """
    jacbn = np.zeros((len(zlist), len(zlist)))
    div = 1
    new = 1000
    old = 2000
    check = 1000
    while check > 1.0e-7:
        for i, _ in enumerate(zlist):
            h_var = abs(zlist[i] * 1.0e-8) / div
            h2_var = 2 * h_var
            zlist[i] = zlist[i] - h_var
            ini_val = model(zlist, sp_comp, ks_dict, r_dict,
                            p_dict, stch_var)
            zlist[i] = zlist[i] + 2 * h_var
            out_val = model(zlist, sp_comp, ks_dict, r_dict,
                            p_dict, stch_var)
            zlist[i] = zlist[i] - h_var
            for j, _ in enumerate(out_val):
                jacbn[j, i] = (out_val[j] - ini_val[j]) / h2_var
        div = div * 2
        old = new
        new = sum([jacbn[k, k] for k in range(len(zlist))])
        check = abs(new - old) / abs(new)
    return np.array(jacbn)


def lna_model_ss(zlist, sp_comp, ks_dict, r_dict, p_dict, stch_var):
    """This  function returns the derivative of components with respect
    to time at a particular state of the system based on the inputs.

    Args:
        zlist (list): list of components or species amounts
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (np.ndarray): stoichiometric matrix

    Returns:
        np.ndarray: derivative of species with respect to time.d(S*f)/dt
    """
    conc = {}
    ind = 0
    for spi in sp_comp:
        conc[spi] = zlist[ind]
        ind = ind + 1
    prop_flux = propensity_vec_molar(
        ks_dict, conc, r_dict, p_dict, True)
    fofx = np.matmul(stch_var, prop_flux)
    return fofx.reshape(len(sp_comp))


def lna_steady_state(t_var, sp_comp, ks_dict, conc, r_dict, p_dict,
                     stch_var, items=None):
    """[summary]

    Args:
        t_var (list): time stamp
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (np.ndarray): stoichiometric matrix

        items (tuplel): (canvas, scroll_x, scroll_y). Defaults to None.

    Returns:
        [type]: [description]
    """
    ind = 0
    zlist = []
    for spi in sp_comp:
        zlist.append(conc[spi])
    zlist = fsolve(
        lna_model_ss,
        tuple(zlist),
        xtol=1.0e-10,
        args=(sp_comp, ks_dict, r_dict, p_dict, stch_var))
    ind = 0
    for spi in sp_comp:
        conc[spi] = zlist[ind]
        ind = ind + 1
    dsf_dx = lna_ss_jacobian(
        lna_model_ss, zlist, sp_comp, stch_var, ks_dict, r_dict, p_dict)
    f_prop = propensity_vec_molar(ks_dict, conc, r_dict, p_dict, True)
    bb_diffsn = np.matmul(
        np.matmul(
            stch_var,
            np.diag(
                f_prop.flatten())),
        stch_var.T)

    dsf_dx = np.nan_to_num(dsf_dx)
    bb_diffsn = np.nan_to_num(bb_diffsn)
    cov_mat = LA.solve_continuous_lyapunov(dsf_dx, -bb_diffsn)

    if items:
        text = prepare_scroll_text(items)

        def fprint(xvar):
            return text.insert('insert', " ".join([str(y) for y in xvar]))
    else:
        def fprint(xvar):
            return print(" ".join([str(y) for y in xvar]), end="")

    fprint(["\nConcentrations\n\n"])
    ind = 0
    for spi in sp_comp:
        if spi[0] != "-":
            fprint([spi, " = ", zlist[ind], "\n"])
            ind = ind + 1

    fprint(["\nCovariance\n\n"])
    i_ind = 0
    for spi in sp_comp:
        j = 0
        for spj in sp_comp:
            if j >= i_ind:
                val = cov_mat[i_ind, j]
                if str(val) not in {"None", "nan", "0.0"} \
                        and spi[0] != "-" and spj[0] != "-":
                    fprint([" ".join(["Covr", spi, spj])
                            .ljust(50), "=", val, "\n"])
            j = j + 1
        i_ind = i_ind + 1
    fprint(["\nCorrelation\n\n"])
    i_ind = 0
    for spi in sp_comp:
        j = 0
        for spj in sp_comp:
            if j >= i_ind:
                val = cov_mat[i_ind, j] / \
                    np.sqrt(np.abs(cov_mat[i_ind, i_ind]
                                   * cov_mat[j, j]))
                if str(val) not in {"None", "nan", "0.0"} \
                        and spi[0] != "-" and spj[0] != "-":
                    fprint([" ".join(["Corr", spi, spj]).ljust(50),
                            "=", val, "\n"])
            j = j + 1
        i_ind = i_ind + 1

    fprint(["\nFano Factor\n\n"])
    ind = 0
    for spi in sp_comp:
        val = cov_mat[ind, ind] / zlist[ind]
        if str(val) not in {"None", "nan", "0.0"} and spi[0] != "-":
            fprint([" ".join(["Fano Factor for", spi])
                    .ljust(50), "=", val, "\n"])
        ind = ind + 1

    return [cov_mat, zlist]



"""

                     This module is the BioSSl module

This module process SSL (structured simulation language) queries.


"""

# import sys
import os
import webbrowser
from sys import platform as PLATFORM
import pandas as pd
import matplotlib.pyplot as plt
# sys.path.append(os.path.abspath("BioSANS2020"))

# import pathlib
from BioSANS2020.prepcodes.process import process
from BioSANS2020.cli_functs.ssl_calls \
    import load_data_traj, calc_average_conc_at_tend, calc_covariance, \
    prob_density_calc_wtime, prob_density_calc  # , \
# calc_covariance_per_traj, calc_covariance_bootsrap, \
# prob_density_calc_tslice
from BioSANS2020.myglobal import mglobals as globals2

if PLATFORM == "win32":
    pass
    # from subprocess import Popen, CREATE_NEW_CONSOLE
elif PLATFORM == "darwin":
    # try:
    # from applescript import tell as my_tell_us
    # except:
    # pass
    from subprocess import call as my_call_us
elif PLATFORM == "linux":
    pass
else:
    pass
    # from subprocess import Popen

if 'USERPROFILE' in os.environ:
    CWD = os.path.join(os.environ['USERPROFILE'], "BioSSL_temporary_folder")
elif 'HOME' in os.environ:
    CWD = os.path.join(os.environ['HOME'], "BioSSL_temporary_folder")
else:
    try:
        import tempfile
        CWD = os.path.join(tempfile.gettempdir(), "BioSSL_temporary_folder")
    except:
        CWD = os.path.join(os.getcwd(), "BioSSL_temporary_folder")

try:
    os.mkdir(CWD, 0o777)
except:
    pass

globals2.init(globals2)
TRAJ = {}


def eval2(to_eval):
    """Evaluate expression"""
    return eval(to_eval)


def show_file_dir(path):
    """This function opens the current working directory

    Args:
        path (str): current working directory
    """
    if PLATFORM == "win32":
        os.startfile(os.path.dirname(path))
    elif PLATFORM == "darwin":
        my_call_us('open', os.path.dirname(path))
    elif PLATFORM == "linux":
        my_call_us('xdg-open', os.path.dirname(path))
    else:
        webbrowser.open(os.path.dirname(path))


def get_input():
    """Thi function get the input commands as user type in console

    Returns:
        str: The input command
    """
    try:
        return input("> ")
    except:
        return raw_input("> ")


def process_command(command):
    """This function handles the processing of commands.

    Args:
        command (str): string of commands
    """
    global CWD
    rowc = command.strip().split()
    miter = 1
    tend = 100
    vol = 1.0
    tsc = 1.5
    tlen = 100
    mult_proc = False
    fout = "temp_traj"
    file_unit = "molecules"
    plot = True
    mixp = True
    norm = False
    logx = False
    logy = False
    edatafile = None
    topo = None
    if rowc[0].lower() == "propagate":
        rlen = len(rowc)
        i = 1
        rxns = ""
        while rowc[i] != "where" and i < rlen:
            rxns = rxns + " " + rowc[i]
            i = i + 1
        rxns = rxns.split("&")
        # try:
        ffile = CWD + "/temp.txt"
        fvar = open(ffile, "w")
        fvar.write("#REACTIONS\n")
        for r_x in rxns:
            fvar.write(r_x + "\n")
        fvar.write("\n")
        mname = ""
        fvar.write("@CONCENTRATION\n")
        conc = ""

        i = i + 1
        while rowc[i] != "using" and i < rlen:
            conc = conc + rowc[i]
            mname = rowc[i + 2]
            i = i + 1

        for c_x in conc.split("&"):
            fvar.write(c_x.replace("=", ",") + "\n")
        fvar.write("\n")
        fvar.close()

        i = i + 3
        opts = ""
        while i < rlen:
            opts = opts + rowc[i]
            i = i + 1
        opts = opts.split("&")
        optsv = {}
        for o_p in opts:
            opr = o_p.split("=")
            try:
                optsv[opr[0].strip()] = opr[1]
            except:
                pass

        if 'tend' in optsv:
            tend = float(optsv['tend'])
        if 'tlen' in optsv:
            tlen = int(optsv['tlen'])
        if 'Vol' in optsv:
            vol = float(optsv['Vol'])
        if 'tsc' in optsv:
            tsc = float(optsv['tsc'])
        if 'miter' in optsv:
            miter = int(optsv['miter'])
        if 'mult_proc' in optsv:
            mult_proc = optsv['mult_proc'].lower() == "true"
        if 'fout' in optsv:
            fout = optsv['fout']
        if 'plot' in optsv:
            plot = optsv['plot'].lower() == "true"
        if 'norm' in optsv:
            norm = optsv['norm'].lower() == "true"
        if 'mixp' in optsv:
            mixp = optsv['mixp'].lower() == "true"
        if 'logx' in optsv:
            norm = optsv['norm'].lower() == "logx"
        if 'logy' in optsv:
            norm = optsv['norm'].lower() == "logy"
        if 'fileUnit' in optsv:
            file_unit = optsv['fileUnit'].lower()
        if 'EdataFile' in optsv:
            edatafile = optsv['EdataFile']
            edatafile = os.path.join(CWD, edatafile)
        if 'topo' in optsv:
            topo = optsv['topo']
            topo = os.path.join(CWD, topo)
            ffile = topo

        fname = CWD + "/" + fout
        process(
            rfile=ffile,
            miter=miter,
            conc_unit=file_unit,
            v_volms=vol,
            tend=tend,
            del_coef=tsc,
            normalize=norm,
            logx=logx,
            logy=logy,
            method=mname,
            tlen=tlen,
            mix_plot=mixp,
            save=True,
            out_fname=fname,
            plot_show=plot,
            c_input={},
            vary="",
            mult_proc=mult_proc,
            items=None,
            exp_data_file=edatafile
        )
        # except:
        #print("temp.txt file not found")
    elif rowc[0].lower() == "load":
        sslfile = CWD + "/" + rowc[1]
        fvar = open(sslfile)
        command = ""
        for row in fvar:
            if len(row.strip()) > 0:
                if row.strip()[-1] == ";":
                    command = command + row.strip().replace(";", "")
                    process_command(command)
                    command = ""
                else:
                    command = command + row.strip() + " "
        fvar.close()
    elif rowc[0].lower() == "pwd":
        print(CWD)
    elif rowc[0].lower() == "mkdir":
        try:
            os.mkdir(CWD + "/" + rowc[1], 0o777)
        except:
            pass
    elif rowc[0].lower() == "ls":
        dirs = []
        try:
            if len(rowc) > 1:
                dirs = os.listdir(CWD + "/" + rowc[1])
            else:
                dirs = os.listdir(CWD)
        except:
            pass
        for xvar in dirs:
            if os.path.isdir(xvar):
                print("directory : " + xvar)
            else:
                print("file      : " + xvar)
    elif rowc[0].lower() == "cd":
        try:
            if os.path.isdir(CWD + "/" + rowc[1]):
                abspath = os.path.abspath(CWD + "/" + rowc[1])
                dname = os.path.dirname(abspath)
                os.chdir(dname)
                CWD = str(abspath)
            elif os.path.isdir(rowc[1].strip()):
                abspath = os.path.abspath(rowc[1])
                dname = os.path.dirname(abspath)
                os.chdir(dname)
                CWD = str(abspath)
            else:
                print("No such directory")
        except:
            print("cannot change dir")
    elif rowc[0].lower() == "read_traj":
        if len(rowc) == 4:
            if rowc[2].lower() == "as":
                name = rowc[3].strip()
                try:
                    sslfile = CWD + "/" + rowc[1]
                    TRAJ[name] = load_data_traj(sslfile)
                except:
                    try:
                        sslfile = rowc[1]
                        TRAJ[name] = load_data_traj(sslfile)
                    except:
                        print("File not found")
            else:
                print("use as to assign content to variable")
                print("read_traj filename as variable")
        else:
            print("Invalid syntax")
    elif rowc[0].lower() == "print":
        try:
            if len(rowc) == 3:
                try:
                    print(TRAJ[rowc[1].strip()][eval2(rowc[2].strip())])
                except:
                    print(TRAJ[rowc[1].strip()]
                          [eval2("'" + rowc[2].strip() + "'")])
            else:
                print(TRAJ[rowc[1].strip()])
        except:
            pass
    elif rowc[0].lower() == "plot":
        if len(rowc) == 4:
            TRAJ[rowc[1].strip()].plot(
                xvar=rowc[2], y=rowc[3], kind='scatter', s=1)
            plt.show()
        elif len(rowc) > 4:
            q_x = rowc[3:]
            plt.xlabel(rowc[2])
            for xvar in q_x:
                plt.scatter(TRAJ[rowc[1]][rowc[2]],
                            TRAJ[rowc[1]][xvar], label=xvar, s=1)
            plt.legend()
            plt.show()
        else:
            print("Invalid syntax")
    elif rowc[0].lower() == "calc_covariance":
        try:
            calc_covariance(TRAJ[rowc[1].strip()], int(rowc[2]))
        except:
            calc_covariance(TRAJ[rowc[1].strip()], 100)
    elif rowc[0].lower() == "prob_density":
        prob_density_calc(TRAJ[rowc[1].strip()], CWD + "/" + rowc[1].strip())
    elif rowc[0].lower() == "prob_density_wtime":
        prob_density_calc_wtime(TRAJ[rowc[1].strip()], CWD +
                                "/" + rowc[1].strip(), "Mname")
    elif rowc[0].lower() == "calc_average":
        try:
            calc_average_conc_at_tend(TRAJ[rowc[1].strip()], int(rowc[2]))
        except:
            calc_average_conc_at_tend(TRAJ[rowc[1].strip()], 100)
    elif rowc[0].lower() == "length":
        try:
            print(len(TRAJ[rowc[1].strip()]))
        except:
            pass
    elif rowc[0].lower() == "pdread_traj":
        if len(rowc) == 4:
            if rowc[2].lower() == "as":
                name = rowc[3].strip()
                try:
                    sslfile = CWD + "/" + rowc[1]
                    TRAJ[name] = pd.read_csv(sslfile, delimiter="\t")
                except:
                    try:
                        sslfile = rowc[1]
                        TRAJ[name] = pd.read_csv(sslfile, delimiter="\t")
                    except:
                        print("file not found")
            else:
                print("use as to assign content to variable")
                print("read_traj filename as variable")
        else:
            print("invalid syntax")
    elif rowc[0].lower() == "open_pwd":
        show_file_dir(CWD + "/.")


if __name__ == '__main__':
    print("###############################################################")
    print("Welcome to BioSSL commandline interface\n")

    ABSPATH = CWD  # os.path.abspath(CWD)
    DNAME = os.path.dirname(ABSPATH)
    os.chdir(DNAME)
    CWD = str(ABSPATH).replace("\\", "/")

    ROW = " "
    COMMAND = " "

    while ROW.strip() != "quit()":
        ROW = " " + get_input().strip()
        if ROW[-1] == ";":
            COMMAND = COMMAND + " " + ROW.strip().replace(";", "")
            if COMMAND.strip() != "":
                process_command(COMMAND)
            COMMAND = " "
            ROW = " "
        else:
            COMMAND = COMMAND + " " + ROW
